from typing import Any, Callable, Dict, Optional, Sequence

from collections import defaultdict
from copy import deepcopy
from datetime import datetime

from anthropic import AsyncMessageStreamManager, AsyncStream, Client, MessageStreamManager, Stream
from anthropic.types import ContentBlockDeltaEvent, InputJSONDelta, Message, MessageDeltaEvent, MessageStartEvent, TextBlock, ToolUseBlock

from parea.cache.cache import Cache
from parea.helpers import timezone_aware_now
from parea.schemas import CacheRequest, LLMInputs, ModelParams
from parea.schemas import Role as PareaRole
from parea.schemas import TraceLog
from parea.utils.trace_utils import make_output, trace_data
from parea.wrapper import Wrapper
from parea.wrapper.anthropic.stream_wrapper import AnthropicAsyncStreamWrapper, AnthropicStreamWrapper, MessageAsyncStreamManagerWrapper, MessageStreamManagerWrapper
from parea.wrapper.utils import _compute_cost


class AnthropicWrapper:

    def init(self, log: Callable, cache: Cache, client: Client):
        func_names = ["messages.create", "messages.stream"]
        if hasattr(client, "beta") and hasattr(client.beta, "tools") and hasattr(client.beta.tools, "messages") and hasattr(client.beta.tools.messages, "create"):
            func_names.append("beta.tools.messages.create")
        Wrapper(
            resolver=self.resolver,
            gen_resolver=self.gen_resolver,
            agen_resolver=self.agen_resolver,
            should_use_gen_resolver=self.should_use_gen_resolver,
            log=log,
            module=client,
            func_names=func_names,
            cache=cache,
            convert_kwargs_to_cache_request=self.convert_kwargs_to_cache_request,
            convert_cache_to_response=self.convert_cache_to_response,
            aconvert_cache_to_response=self.aconvert_cache_to_response,
            name="llm-anthropic",
        )

    @staticmethod
    def resolver(trace_id: str, _args: Sequence[Any], kwargs: Dict[str, Any], response: Optional[Message]) -> Optional[Any]:
        if response:
            if len(response.content) > 1:
                output_list = []
                for content in response.content:
                    if isinstance(content, TextBlock):
                        output_list.append(content.text)
                    elif isinstance(content, ToolUseBlock):
                        output_list.append(content.model_dump())
                output = make_output(output_list, islist=True)
            else:
                output = response.content[0].text
            input_tokens = response.usage.input_tokens
            output_tokens = response.usage.output_tokens
            total_tokens = input_tokens + output_tokens
        else:
            output = None
            input_tokens, output_tokens, total_tokens = 0, 0, 0

        llm_configuration = AnthropicWrapper._kwargs_to_llm_configuration(kwargs)
        model = response.model if response and response.model else llm_configuration.model

        trace_data.get()[trace_id].configuration = llm_configuration
        trace_data.get()[trace_id].input_tokens = input_tokens
        trace_data.get()[trace_id].output_tokens = output_tokens
        trace_data.get()[trace_id].total_tokens = total_tokens
        trace_data.get()[trace_id].cost = _compute_cost(input_tokens, output_tokens, model)
        trace_data.get()[trace_id].output = output
        return response

    def gen_resolver(self, trace_id: str, _args: Sequence[Any], kwargs: Dict[str, Any], response, final_log):
        if isinstance(response, Stream):
            llm_configuration = self._kwargs_to_llm_configuration(kwargs)
            trace_data.get()[trace_id].configuration = llm_configuration
            accumulator, info_from_response = self._get_default_dict_streaming()

            def gen_final_processing_and_logging(accumulator, info_from_response):
                model = info_from_response.get("model") or llm_configuration.model
                self.update_trace_data_from_stream_response(trace_id, model, accumulator, info_from_response)
                final_log()

            return AnthropicStreamWrapper(
                response,
                accumulator,
                info_from_response,
                self._update_accumulator_streaming,
                gen_final_processing_and_logging,
            )
        else:  # needed for stream context manager; this will need to handle both sync and async; as messages.stream is always sync

            def resolve_and_log(m: Message):
                self.resolver(trace_id, _args, kwargs, m)
                final_log()

            if isinstance(response, MessageStreamManager):
                return MessageStreamManagerWrapper(response, resolve_and_log)
            else:
                return MessageAsyncStreamManagerWrapper(response, resolve_and_log)

    def agen_resolver(self, trace_id: str, _args: Sequence[Any], kwargs: Dict[str, Any], response, final_log):
        llm_configuration = self._kwargs_to_llm_configuration(kwargs)
        trace_data.get()[trace_id].configuration = llm_configuration
        accumulator, info_from_response = self._get_default_dict_streaming()

        def gen_final_processing_and_logging(accumulator, info_from_response):
            model = info_from_response.get("model") or llm_configuration.model
            self.update_trace_data_from_stream_response(trace_id, model, accumulator, info_from_response)
            final_log()

        return AnthropicAsyncStreamWrapper(
            response,
            accumulator,
            info_from_response,
            self._update_accumulator_streaming,
            gen_final_processing_and_logging,
        )

    @staticmethod
    def _get_default_dict_streaming():
        info_from_response = {"model": "", "first_token_timestamp": None, "input_tokens": None, "output_tokens": None}
        accumulator = defaultdict(content=[], role="assistant")
        return accumulator, info_from_response

    @staticmethod
    def _kwargs_to_llm_configuration(kwargs, model=None) -> LLMInputs:
        functions = deepcopy([d for d in kwargs.get("tools", [])])
        messages = kwargs.get("messages", None)
        if system_msg := kwargs.get("system", None):
            if not messages:
                messages = []
            messages = messages.copy()
            messages.insert(0, dict(role=PareaRole.system, content=system_msg))
        for func in functions:
            if "input_schema" in func:
                func["parameters"] = func.pop("input_schema")
        return LLMInputs(
            model=model or kwargs.get("model", None),
            provider="anthropic",
            messages=messages,
            model_params=ModelParams(
                temp=kwargs.get("temperature", 1.0),
                max_length=kwargs.get("max_tokens", None),
                top_p=kwargs.get("top_p", 1.0),
            ),
            functions=functions,
        )

    def convert_kwargs_to_cache_request(self, _args: Sequence[Any], kwargs: Dict[str, Any]) -> CacheRequest:
        pass

    @staticmethod
    def update_trace_data_from_stream_response(trace_id, model, accumulator, model_info):
        output = accumulator["content"]
        if isinstance(output, list):
            output = "".join(output)
        trace_data.get()[trace_id].output = output
        output_tokens = model_info.get("output_tokens", 0)
        input_tokens = model_info.get("input_tokens", 0)
        trace_data.get()[trace_id].input_tokens = input_tokens
        trace_data.get()[trace_id].output_tokens = output_tokens
        trace_data.get()[trace_id].total_tokens = input_tokens + output_tokens
        trace_data.get()[trace_id].cost = _compute_cost(input_tokens, output_tokens, model)
        trace_data.get()[trace_id].time_to_first_token = (model_info["first_token_timestamp"] - datetime.fromisoformat(trace_data.get()[trace_id].start_timestamp)).total_seconds()

    @staticmethod
    def _convert_cache_to_response(_args: Sequence[Any], kwargs: Dict[str, Any], cache_response: TraceLog):
        pass

    @staticmethod
    def should_use_gen_resolver(response: Any) -> bool:
        return isinstance(response, (Stream, MessageStreamManager, AsyncStream, AsyncMessageStreamManager))

    def convert_cache_to_response(self, _args: Sequence[Any], kwargs: Dict[str, Any], cache_response: TraceLog):
        pass

    def aconvert_cache_to_response(self, _args: Sequence[Any], kwargs: Dict[str, Any], cache_response: TraceLog):
        pass

    @staticmethod
    def _update_accumulator_streaming(accumulator, info_from_response, chunk):
        if isinstance(chunk, MessageStartEvent):
            info_from_response["input_tokens"] = chunk.message.usage.input_tokens
        elif isinstance(chunk, ContentBlockDeltaEvent):
            if isinstance(chunk.delta, InputJSONDelta):
                accumulator["content"].append(chunk.delta.partial_json)
            else:
                accumulator["content"].append(chunk.delta.text)
            if not info_from_response.get("first_token_timestamp"):
                info_from_response["first_token_timestamp"] = timezone_aware_now()
        elif isinstance(chunk, MessageDeltaEvent):
            info_from_response["output_tokens"] = chunk.usage.output_tokens



from .answer_context_faithfulness_binary import answer_context_faithfulness_binary_factory
from .answer_context_faithfulness_precision import answer_context_faithfulness_precision_factory
from .answer_context_faithfulness_statement_level import answer_context_faithfulness_statement_level_factory
from .context_has_answer import context_has_answer_factory
from .context_query_relevancy import context_query_relevancy_factory
from .context_ranking_listwise import context_ranking_listwise_factory
from .context_ranking_pointwise import context_ranking_pointwise_factory
from .percent_target_supported_by_context import percent_target_supported_by_context_factory



from __future__ import annotations

from types import TracebackType
from typing import Callable

from anthropic import AsyncMessageStreamManager, AsyncStream, MessageStreamManager, Stream
from anthropic.types import Message


class AnthropicStreamWrapper:
    def __init__(self, stream: Stream, accumulator, info_from_response, update_accumulator_streaming, final_processing_and_logging):
        self._stream = stream
        self._final_processing_and_logging = final_processing_and_logging
        self._update_accumulator_streaming = update_accumulator_streaming
        self._accumulator = accumulator
        self._info_from_response = info_from_response

    def __getattr__(self, attr):
        # delegate attribute access to the original stream
        return getattr(self._stream, attr) if hasattr(self._stream, attr) else None

    def __iter__(self):
        for chunk in self._stream:
            self._update_accumulator_streaming(self._accumulator, self._info_from_response, chunk)
            yield chunk

        self._final_processing_and_logging(self._accumulator, self._info_from_response)


class AnthropicAsyncStreamWrapper:
    def __init__(self, stream: AsyncStream, accumulator, info_from_response, update_accumulator_streaming, final_processing_and_logging):
        self._stream = stream
        self._final_processing_and_logging = final_processing_and_logging
        self._update_accumulator_streaming = update_accumulator_streaming
        self._accumulator = accumulator
        self._info_from_response = info_from_response

    def __getattr__(self, attr):
        # delegate attribute access to the original async_stream
        return getattr(self._stream, attr) if hasattr(self._stream, attr) else None

    async def __aiter__(self):
        async for chunk in self._stream:
            self._update_accumulator_streaming(self._accumulator, self._info_from_response, chunk)
            yield chunk

        self._final_processing_and_logging(self._accumulator, self._info_from_response)


class MessageStreamManagerWrapper(MessageStreamManager):
    def __init__(self, msm_instance: MessageStreamManager, resolve_and_log: Callable):
        self._msm_instance = msm_instance
        self._resolve_and_log = resolve_and_log

    def __getattr__(self, attr):
        if attr != "_private_stream":
            return getattr(self._msm_instance, attr)
        else:
            return self._private_stream

    def __enter__(self):
        self._private_stream = self._msm_instance.__enter__()
        return self._private_stream

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        m: Message = self._private_stream.get_final_message()
        self._resolve_and_log(m)
        return super().__exit__(exc_type, exc, exc_tb)


class MessageAsyncStreamManagerWrapper(AsyncMessageStreamManager):
    def __init__(self, msm_instance: AsyncMessageStreamManager, resolve_and_log: Callable):
        self._msm_instance = msm_instance
        self._resolve_and_log = resolve_and_log

    def __getattr__(self, attr):
        if attr != "_private_stream":
            return getattr(self._msm_instance, attr)
        else:
            return self._private_stream

    async def __aenter__(self):
        self._private_stream = await self._msm_instance.__aenter__()
        return self._private_stream

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        m: Message = await self._private_stream.get_final_message()
        self._resolve_and_log(m)
        return await super().__aexit__(exc_type, exc, exc_tb)



from typing import Union

from Levenshtein import distance

from parea.schemas import Log


def levenshtein(log: Log) -> Union[float, None]:
    output = log.output
    if (target := log.target) is None:
        return None

    return levenshtein_distance(str(output), str(target))


def levenshtein_distance(output: str, target: str) -> float:
    max_len = max(len(x) for x in [output, target])

    score = 1
    if max_len > 0:
        score = 1 - (distance(output, target) / max_len)

    return score



from typing import Any, AsyncGenerator, AsyncIterator, Dict, Generator, Iterator

import contextvars
import inspect
import json
import logging
import os
from datetime import datetime

from openai import AsyncOpenAI
from openai.types.beta.threads import Run

from parea.constants import PAREA_OS_ENV_EXPERIMENT_UUID
from parea.helpers import gen_trace_id, is_logging_disabled, timezone_aware_now
from parea.schemas import TraceLog, UpdateTraceScenario
from parea.utils.trace_utils import check_multiple_return_values, logger_all_possible, make_output, trace_context, trace_data
from parea.utils.universal_encoder import json_dumps
from parea.wrapper.utils import _compute_cost

logger = logging.getLogger()


class BaseWrapper:
    @staticmethod
    def fill_trace_data(trace_id: str, data: Dict[str, Any], scenario: UpdateTraceScenario):
        try:
            if scenario == UpdateTraceScenario.RESULT:
                if not isinstance(data["result"], (Generator, AsyncGenerator, AsyncIterator, Iterator)):
                    trace_data.get()[trace_id].output = make_output(data["result"], data.get("output_as_list", False))
                trace_data.get()[trace_id].status = "success"
                trace_data.get()[trace_id].evaluation_metric_names = data.get("eval_funcs_names")
            elif scenario == UpdateTraceScenario.ERROR:
                trace_data.get()[trace_id].error = data["error"]
                trace_data.get()[trace_id].status = "error"
            elif scenario == UpdateTraceScenario.CHAIN:
                trace_data.get()[trace_id].parent_trace_id = data["parent_trace_id"]
                trace_data.get()[data["parent_trace_id"]].children.append(trace_id)
            elif scenario == UpdateTraceScenario.USAGE:
                if usage := data.get("usage", {}):
                    if usage:
                        trace_data.get()[trace_id].input_tokens = usage.get("prompt_tokens", 0)
                        trace_data.get()[trace_id].output_tokens = usage.get("completion_tokens", 0)
                        trace_data.get()[trace_id].total_tokens = usage.get("total_tokens", 0)
                        if data.get("model", None):
                            trace_data.get()[trace_id].cost = _compute_cost(usage.get("prompt_tokens", 0), usage.get("total_tokens", 0), data.get("model", None))
            else:
                logger.debug(f"Error occurred filling trace data. Scenario not valid: {scenario}")
        except Exception as e:
            logger.debug(
                f"Error occurred filling trace data for trace id {trace_id}, {e}",
                exc_info=e,
            )

    def init_trace(self, func_name, args, kwargs, func):
        start_time = timezone_aware_now()
        trace_id = gen_trace_id()

        new_trace_context = trace_context.get() + [trace_id]
        token = trace_context.set(new_trace_context)

        if is_logging_disabled():
            return trace_id, start_time, token

        try:
            sig = inspect.signature(func)
            parameters = sig.parameters

            inputs = {k: v for k, v in zip(parameters.keys(), args)}
            inputs.update(kwargs)

            # filter out any values which aren't JSON serializable
            for k, v in inputs.items():
                try:
                    json.dumps(v)
                except TypeError:
                    try:
                        inputs[k] = json_dumps(v)
                    except TypeError:
                        # if we can't serialize the value, just convert it to a string
                        inputs[k] = str(v)

            trace_data.get()[trace_id] = TraceLog(
                trace_id=trace_id,
                parent_trace_id=trace_id,
                root_trace_id=new_trace_context[0],
                start_timestamp=start_time.isoformat(),
                trace_name=func_name,
                inputs=inputs,
                experiment_uuid=os.environ.get(PAREA_OS_ENV_EXPERIMENT_UUID, None),
            )

            parent_trace_id = new_trace_context[-2] if len(new_trace_context) > 1 else None
            if parent_trace_id:
                self.fill_trace_data(
                    trace_id,
                    {"parent_trace_id": parent_trace_id},
                    UpdateTraceScenario.CHAIN,
                )
        except Exception as e:
            logger.debug(f"Error occurred initializing trace for function {func_name}, {e}")

        return trace_id, start_time, token

    @staticmethod
    def cleanup_trace(trace_id: str, start_time: datetime, context_token: contextvars.Token):
        end_time = timezone_aware_now()
        trace_data.get()[trace_id].end_timestamp = end_time.isoformat()
        trace_data.get()[trace_id].latency = (end_time - start_time).total_seconds()

        output = trace_data.get()[trace_id].output
        if trace_data.get()[trace_id].status == "success" and output:
            output_for_eval_metrics = output
            trace_data.get()[trace_id].output_for_eval_metrics = output_for_eval_metrics

        logger_all_possible(trace_id)
        trace_context.reset(context_token)

    def _wrap_steps(self, module_name, func, *args, **kwargs):
        trace_id, start_time, context_token = self.init_trace(f"{module_name}.{func.__name__}", args, kwargs, func)
        output_as_list = check_multiple_return_values(func)
        try:
            result = func(*args, **kwargs)
            self.fill_trace_data(
                trace_id,
                {
                    "result": result,
                    "output_as_list": output_as_list,
                },
                UpdateTraceScenario.RESULT,
            )
            if isinstance(result, Run) and result.usage:
                self.fill_trace_data(
                    trace_id,
                    result.model_dump(),
                    UpdateTraceScenario.USAGE,
                )
            return result
        except Exception as e:
            logger.error(f"Error occurred in function {func.__name__}, {e}")
            self.fill_trace_data(trace_id, {"error": str(e)}, UpdateTraceScenario.ERROR)
            raise e
        finally:
            try:
                self.cleanup_trace(trace_id, start_time, context_token)
            except Exception as e:
                logger.debug(
                    f"Error occurred cleaning up trace for function {func.__name__}, {e}",
                    exc_info=e,
                )

    async def _awrap_steps(self, module_name, func, *args, **kwargs):
        trace_id, start_time, context_token = self.init_trace(f"{module_name}.{func.__name__}", args, kwargs, func)
        output_as_list = check_multiple_return_values(func)
        try:
            result = await func(*args, **kwargs)
            self.fill_trace_data(
                trace_id,
                {
                    "result": result,
                    "output_as_list": output_as_list,
                },
                UpdateTraceScenario.RESULT,
            )
            if isinstance(result, Run) and result.usage:
                self.fill_trace_data(
                    trace_id,
                    result.model_dump(),
                    UpdateTraceScenario.USAGE,
                )
            return result
        except Exception as e:
            logger.error(f"Error occurred in function {func.__name__}, {e}")
            self.fill_trace_data(trace_id, {"error": str(e)}, UpdateTraceScenario.ERROR)
            raise e
        finally:
            try:
                self.cleanup_trace(trace_id, start_time, context_token)
            except Exception as e:
                logger.debug(
                    f"Error occurred cleaning up trace for function {func.__name__}, {e}",
                    exc_info=e,
                )


class AssistantsWrapper(BaseWrapper):
    def __init__(self, client, is_async):
        self.module_name = "beta.assistants"
        self.client = client
        self.wrapped = False
        self.is_async = is_async
        try:
            self.create_method = client.beta.assistants.create
            self.retrieve_method = client.beta.assistants.retrieve
            self.update_method = client.beta.assistants.update
            self.list_method = client.beta.assistants.list
            self.delete_method = client.beta.assistants.delete
            self.wrapped = True
        except Exception as e:
            logger.debug(f"Error occurred initializing AssistantsWrapper, {e}")

    def create(self, *args, **kwargs):
        func = self.create_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def retrieve(self, *args, **kwargs):
        func = self.retrieve_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def update(self, *args, **kwargs):
        func = self.update_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def list(self, *args, **kwargs):
        func = self.list_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def delete(self, *args, **kwargs):
        func = self.delete_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    async def acreate(self, *args, **kwargs):
        func = self.create_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def aretrieve(self, *args, **kwargs):
        func = self.retrieve_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def aupdate(self, *args, **kwargs):
        func = self.update_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def alist(self, *args, **kwargs):
        func = self.list_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def adelete(self, *args, **kwargs):
        func = self.delete_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    def init(self):
        if not self.wrapped:
            return

        methods = {
            self.create_method: self.acreate if self.is_async else self.create,
            self.retrieve_method: self.aretrieve if self.is_async else self.retrieve,
            self.update_method: self.aupdate if self.is_async else self.update,
            self.list_method: self.alist if self.is_async else self.list,
            self.delete_method: self.adelete if self.is_async else self.delete,
        }

        for method, function in methods.items():
            setattr(self.client.beta.assistants, method.__name__, function)


class ThreadsWrapper(BaseWrapper):
    def __init__(self, client, is_async):
        self.module_name = "beta.threads"
        self.client = client
        self.wrapped = False
        self.is_async = is_async
        try:
            self.create_method = client.beta.threads.create
            self.retrieve_method = client.beta.threads.retrieve
            self.update_method = client.beta.threads.update
            self.create_and_run_method = client.beta.threads.create_and_run
            self.delete_method = client.beta.threads.delete
            self.wrapped = True
        except Exception as e:
            logger.debug(f"Error occurred initializing ThreadsWrapper, {e}")

    def create(self, *args, **kwargs):
        func = self.create_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def retrieve(self, *args, **kwargs):
        func = self.retrieve_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def update(self, *args, **kwargs):
        func = self.update_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def create_and_run(self, *args, **kwargs):
        if kwargs.get("stream", False):
            logger.debug("Stream is not supported for create_and_run method")
            return None
        func = self.create_and_run_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def delete(self, *args, **kwargs):
        func = self.delete_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    async def acreate(self, *args, **kwargs):
        func = self.create_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def aretrieve(self, *args, **kwargs):
        func = self.retrieve_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def aupdate(self, *args, **kwargs):
        func = self.update_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def acreate_and_run(self, *args, **kwargs):
        if kwargs.get("stream", False):
            logger.debug("Stream is not supported for create_and_run method")
            return None
        func = self.create_and_run_method
        return self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def adelete(self, *args, **kwargs):
        func = self.delete_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    def init(self):
        if not self.wrapped:
            return

        methods = {
            self.create_method: self.acreate if self.is_async else self.create,
            self.retrieve_method: self.aretrieve if self.is_async else self.retrieve,
            self.update_method: self.aupdate if self.is_async else self.update,
            self.create_and_run_method: self.acreate_and_run if self.is_async else self.create_and_run,
            self.delete_method: self.adelete if self.is_async else self.delete,
        }

        for method, function in methods.items():
            setattr(self.client.beta.threads, method.__name__, function)


class ThreadRunsWrapper(BaseWrapper):
    def __init__(self, client, is_async):
        self.module_name = "beta.threads.runs"
        self.client = client
        self.wrapped = False
        self.is_async = is_async
        try:
            self.create_method = client.beta.threads.runs.create
            self.retrieve_method = client.beta.threads.runs.retrieve
            self.update_method = client.beta.threads.runs.update
            self.list_method = client.beta.threads.runs.list
            self.submit_tool_outputs_method = client.beta.threads.runs.submit_tool_outputs
            self.cancel_method = client.beta.threads.runs.cancel
            self.wrapped = True
        except Exception as e:
            logger.debug(f"Error occurred initializing ThreadRunsWrapper, {e}")

    def create(self, *args, **kwargs):
        if kwargs.get("stream", False):
            logger.debug("Stream is not supported for create method")
            return None
        func = self.create_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def retrieve(self, *args, **kwargs):
        func = self.retrieve_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def update(self, *args, **kwargs):
        func = self.update_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def list(self, *args, **kwargs):
        func = self.list_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def cancel(self, *args, **kwargs):
        func = self.cancel_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def submit_tool_outputs(self, *args, **kwargs):
        if kwargs.get("stream", False):
            logger.debug("Stream is not supported for submit_tool_outputs method")
            return None
        func = self.submit_tool_outputs_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    async def acreate(self, *args, **kwargs):
        if kwargs.get("stream", False):
            logger.debug("Stream is not supported for create method")
            return None
        func = self.create_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def aretrieve(self, *args, **kwargs):
        func = self.retrieve_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def aupdate(self, *args, **kwargs):
        func = self.update_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def alist(self, *args, **kwargs):
        func = self.list_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def acancel(self, *args, **kwargs):
        func = self.cancel_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def asubmit_tool_outputs(self, *args, **kwargs):
        if kwargs.get("stream", False):
            logger.debug("Stream is not supported for submit_tool_outputs method")
            return None
        func = self.submit_tool_outputs_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    def init(self):
        if not self.wrapped:
            return

        methods = {
            self.create_method: self.acreate if self.is_async else self.create,
            self.retrieve_method: self.aretrieve if self.is_async else self.retrieve,
            self.update_method: self.aupdate if self.is_async else self.update,
            self.list_method: self.alist if self.is_async else self.list,
            self.cancel_method: self.acancel if self.is_async else self.cancel,
            self.submit_tool_outputs_method: self.asubmit_tool_outputs if self.is_async else self.submit_tool_outputs,
        }

        for method, function in methods.items():
            setattr(self.client.beta.threads.runs, method.__name__, function)


class ThreadMessagesWrapper(BaseWrapper):
    def __init__(self, client, is_async):
        self.module_name = "beta.threads.messages"
        self.client = client
        self.wrapped = False
        self.is_async = is_async
        try:
            self.create_method = client.beta.threads.messages.create
            self.retrieve_method = client.beta.threads.messages.retrieve
            self.update_method = client.beta.threads.messages.update
            self.list_method = client.beta.threads.messages.list
            self.wrapped = True
        except Exception as e:
            logger.debug(f"Error occurred initializing ThreadMessagesWrapper, {e}")

    def create(self, *args, **kwargs):
        func = self.create_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def retrieve(self, *args, **kwargs):
        func = self.retrieve_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def update(self, *args, **kwargs):
        func = self.update_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def list(self, *args, **kwargs):
        func = self.list_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    async def acreate(self, *args, **kwargs):
        func = self.create_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def aretrieve(self, *args, **kwargs):
        func = self.retrieve_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def aupdate(self, *args, **kwargs):
        func = self.update_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def alist(self, *args, **kwargs):
        func = self.list_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    def init(self):
        if not self.wrapped:
            return

        methods = {
            self.create_method: self.acreate if self.is_async else self.create,
            self.retrieve_method: self.aretrieve if self.is_async else self.retrieve,
            self.update_method: self.aupdate if self.is_async else self.update,
            self.list_method: self.alist if self.is_async else self.list,
        }

        for method, function in methods.items():
            setattr(self.client.beta.threads.messages, method.__name__, function)


class ThreadRunsStepsWrapper(BaseWrapper):
    def __init__(self, client, is_async):
        self.module_name = "beta.threads.runs.steps"
        self.client = client
        self.wrapped = False
        self.is_async = is_async
        try:
            self.retrieve_method = client.beta.threads.runs.steps.retrieve
            self.list_method = client.beta.threads.runs.steps.list
            self.wrapped = True
        except Exception as e:
            logger.debug(f"Error occurred initializing ThreadRunsStepsWrapper, {e}")

    def retrieve(self, *args, **kwargs):
        func = self.retrieve_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def list(self, *args, **kwargs):
        func = self.list_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    async def aretrieve(self, *args, **kwargs):
        func = self.retrieve_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def alist(self, *args, **kwargs):
        func = self.list_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    def init(self):
        if not self.wrapped:
            return

        methods = {
            self.retrieve_method: self.aretrieve if self.is_async else self.retrieve,
            self.list_method: self.alist if self.is_async else self.list,
        }

        for method, function in methods.items():
            setattr(self.client.beta.threads.runs.steps, method.__name__, function)


class ThreadMessagesFilesWrapper(BaseWrapper):
    def __init__(self, client, is_async):
        self.module_name = "beta.threads.messages.files"
        self.client = client
        self.wrapped = False
        self.is_async = is_async
        try:
            self.retrieve_method = client.beta.threads.messages.files.retrieve
            self.list_method = client.beta.threads.messages.files.list
            self.wrapped = True
        except Exception as e:
            logger.debug(f"Error occurred initializing ThreadMessagesFilesWrapper, {e}")

    def retrieve(self, *args, **kwargs):
        func = self.retrieve_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    def list(self, *args, **kwargs):
        func = self.list_method
        return self._wrap_steps(self.module_name, func, *args, **kwargs)

    async def aretrieve(self, *args, **kwargs):
        func = self.retrieve_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    async def alist(self, *args, **kwargs):
        func = self.list_method
        return await self._awrap_steps(self.module_name, func, *args, **kwargs)

    def init(self):
        if not self.wrapped:
            return

        methods = {
            self.retrieve_method: self.aretrieve if self.is_async else self.retrieve,
            self.list_method: self.alist if self.is_async else self.list,
        }

        for method, function in methods.items():
            setattr(self.client.beta.threads.messages.files, method.__name__, function)


class BetaWrappers:
    def __init__(self, client):
        self.is_async = isinstance(client, AsyncOpenAI)
        self.client = client

    def init(self):
        aw = AssistantsWrapper(self.client, self.is_async)
        tw = ThreadsWrapper(self.client, self.is_async)
        trw = ThreadRunsWrapper(self.client, self.is_async)
        trsw = ThreadRunsStepsWrapper(self.client, self.is_async)
        tmw = ThreadMessagesWrapper(self.client, self.is_async)
        tmfw = ThreadMessagesFilesWrapper(self.client, self.is_async)
        aw.init()
        tw.init()
        trw.init()
        trsw.init()
        tmw.init()
        tmfw.init()



from typing import Any, AsyncGenerator, AsyncIterator, Callable, Dict, Generator, Iterator, Optional, Sequence, TypeVar, Union

import json
import os
from collections import defaultdict
from datetime import datetime

import openai
from openai import __version__ as openai_version

from parea.helpers import timezone_aware_now
from parea.utils.universal_encoder import json_dumps
from parea.wrapper.utils import _calculate_input_tokens, _compute_cost, _format_function_call, _kwargs_to_llm_configuration, _num_tokens_from_string

if openai_version.startswith("0."):
    from openai.openai_object import OpenAIObject
    from openai.util import convert_to_openai_object
else:
    from openai.types.chat import ChatCompletion as OpenAIObject

    def convert_to_openai_object(kwargs) -> OpenAIObject:
        if "id" not in kwargs:
            kwargs["id"] = "0"
        if "created" not in kwargs:
            kwargs["created"] = 0
        if "object" not in kwargs:
            kwargs["object"] = "chat.completion"
        if "model" not in kwargs:
            kwargs["model"] = "model"
        if "choices" in kwargs and isinstance(kwargs["choices"], list) and len(kwargs["choices"]) > 0:
            if "finish_reason" not in kwargs["choices"][0]:
                kwargs["choices"][0]["finish_reason"] = "stop"
        return OpenAIObject(**kwargs)


from dotenv import load_dotenv

from parea.cache.cache import Cache
from parea.schemas.models import CacheRequest, TraceLog
from parea.utils.trace_utils import trace_data
from parea.wrapper.wrapper import Wrapper

load_dotenv()

is_old_openai = openai_version.startswith("0.")

_T = TypeVar("_T")


class OpenAIWrapper:
    openai.api_key = os.getenv("OPENAI_API_KEY")

    def get_original_methods(self, module_client=openai):
        if is_old_openai:
            original_methods = {"ChatCompletion.create": module_client.ChatCompletion.create, "ChatCompletion.acreate": module_client.ChatCompletion.acreate}
        else:
            try:
                original_methods = {"chat.completions.create": module_client.chat.completions.create}
            except openai.OpenAIError:
                original_methods = {}

            try:
                latest_methods = {"beta.chat.completions.parse": module_client.beta.chat.completions.parse}
                original_methods.update(latest_methods)
            except Exception:
                pass

        return list(original_methods.keys())

    def init(self, log: Callable, cache: Cache = None, module_client=openai):
        Wrapper(
            resolver=self.resolver,
            gen_resolver=self.gen_resolver,
            agen_resolver=self.agen_resolver,
            should_use_gen_resolver=self.should_use_gen_resolver,
            log=log,
            module=module_client,
            func_names=self.get_original_methods(module_client),
            cache=cache,
            convert_kwargs_to_cache_request=self.convert_kwargs_to_cache_request,
            convert_cache_to_response=self.convert_cache_to_response,
            aconvert_cache_to_response=self.aconvert_cache_to_response,
            name="llm-openai",
        )

    def resolver(self, trace_id: str, _args: Sequence[Any], kwargs: Dict[str, Any], response: Optional[Any]) -> Optional[Any]:
        if response:
            output = self._get_output(response)
            if openai_version.startswith("0."):
                usage = response["usage"]
                input_tokens = usage.get("prompt_tokens", 0)
                output_tokens = usage.get("completion_tokens", 0)
                total_tokens = usage.get("total_tokens", 0)
            else:
                usage = response.usage
                input_tokens = usage.prompt_tokens
                output_tokens = usage.completion_tokens
                total_tokens = usage.total_tokens
        else:
            output = None
            input_tokens = 0
            output_tokens = 0
            total_tokens = 0

        llm_configuration = self._kwargs_to_llm_configuration(kwargs)
        model = response.model if response and response.model else llm_configuration.model

        trace_data.get()[trace_id].configuration = llm_configuration
        trace_data.get()[trace_id].input_tokens = input_tokens
        trace_data.get()[trace_id].output_tokens = output_tokens
        trace_data.get()[trace_id].total_tokens = total_tokens
        trace_data.get()[trace_id].cost = _compute_cost(input_tokens, output_tokens, model)
        trace_data.get()[trace_id].output = json_dumps(output) if not isinstance(output, str) else output
        return response

    def gen_resolver(self, trace_id: str, _args: Sequence[Any], kwargs: Dict[str, Any], response, final_log):
        llm_configuration = self._kwargs_to_llm_configuration(kwargs)
        trace_data.get()[trace_id].configuration = llm_configuration

        accumulator, info_from_response = self._get_default_dict_streaming()

        def gen_final_processing_and_logging(accumulator, info_from_response):
            OpenAIWrapper._format_accumulator_in_place(accumulator)

            model = info_from_response.get("model") or llm_configuration.model
            OpenAIWrapper.update_trace_data_from_stream_response(trace_id, model, accumulator, kwargs, info_from_response["first_token_timestamp"])
            final_log()

        if is_old_openai:

            def logging_sync_generator(sync_gen: Generator[_T]) -> Generator[_T]:
                for chunk in sync_gen:
                    self._update_accumulator_streaming(accumulator, info_from_response, chunk)
                    yield chunk
                gen_final_processing_and_logging(accumulator, info_from_response)

            return logging_sync_generator(response)

        else:
            from parea.types import OpenAIStreamWrapper

            return OpenAIStreamWrapper(
                response,
                accumulator,
                info_from_response,
                OpenAIWrapper._update_accumulator_streaming,
                gen_final_processing_and_logging,
            )

    @staticmethod
    def _get_default_dict_streaming():
        info_from_response = {"model": "", "first_token_timestamp": None}
        accumulated_tools = defaultdict(lambda: {"id": "", "function": {"arguments": [], "name": ""}, "type": "function"})
        accumulator = defaultdict(content=[], role="assistant", function_call=defaultdict(arguments=[], name=""), tool_calls=accumulated_tools)
        return accumulator, info_from_response

    @staticmethod
    def _format_accumulator_in_place(accumulator):
        content = "".join(accumulator.get("content"))
        accumulator["content"] = content

        tool_calls = []
        if accumulated_tools := accumulator.get("tool_calls", {}):
            for tool in accumulated_tools.values():
                tool["function"]["arguments"] = "".join(tool["function"]["arguments"])

            tool_calls = list(accumulated_tools.values())
            for tool in tool_calls:
                tool["function"]["arguments"] = json_dumps(json.loads(tool["function"]["arguments"]), indent=2)
        accumulator["tool_calls"] = tool_calls

        if function_call := accumulator.get("function_call"):
            if function_call.get("name"):
                function_call["arguments"] = json_dumps(json.loads("".join(function_call["arguments"])), indent=2)
            else:
                function_call["arguments"] = ""

        return accumulator

    @staticmethod
    def _update_accumulator_streaming(accumulator, info_from_response, chunk):
        def _set_timestamp_if_not_set():
            if not info_from_response.get("first_token_timestamp"):
                info_from_response["first_token_timestamp"] = timezone_aware_now()

        if chunk.model and not info_from_response.get("model"):
            info_from_response["model"] = chunk.model

        if not chunk.choices:
            return

        if is_old_openai:
            delta_dict = chunk.choices[0].delta._previous
        else:
            delta_dict = chunk.choices[0].delta.model_dump()

        if not accumulator["role"]:
            accumulator["role"] = delta_dict.get("role")

        if delta_dict.get("content"):
            _set_timestamp_if_not_set()
            accumulator["content"].append(delta_dict["content"])

        if delta_dict.get("function_call"):
            _set_timestamp_if_not_set()
            accumulator["function_call"]["name"] = delta_dict.get("function_call")["name"] or accumulator["function_call"]["name"]
            if delta_dict.get("function_call", {}).get("arguments"):
                accumulator["function_call"]["arguments"].append(delta_dict["function_call"]["arguments"])

        if tool_calls := delta_dict.get("tool_calls", []):
            _set_timestamp_if_not_set()
            for tool_call in tool_calls:
                tool_id = tool_call["index"]

                accumulator["tool_calls"][tool_id]["id"] = tool_call.get("id") or accumulator["tool_calls"][tool_id]["id"]
                if not accumulator["tool_calls"][tool_id]["type"]:
                    accumulator["tool_calls"][tool_id]["type"] = "function"

                accumulator["tool_calls"][tool_id]["function"]["name"] = tool_call.get("function", {}).get("name") or accumulator["tool_calls"][tool_id]["function"]["name"]
                if tool_call.get("function", {}).get("arguments"):
                    accumulator["tool_calls"][tool_id]["function"]["arguments"].append(tool_call["function"]["arguments"])

    def agen_resolver(self, trace_id: str, _args: Sequence[Any], kwargs: Dict[str, Any], response, final_log):
        llm_configuration = self._kwargs_to_llm_configuration(kwargs)
        trace_data.get()[trace_id].configuration = llm_configuration

        accumulator, info_from_response = self._get_default_dict_streaming()

        def agen_final_processing_and_logging(accumulator, info_from_response):
            OpenAIWrapper._format_accumulator_in_place(accumulator)

            model = info_from_response.get("model") or llm_configuration.model
            OpenAIWrapper.update_trace_data_from_stream_response(trace_id, model, accumulator, kwargs, info_from_response["first_token_timestamp"])
            final_log()

        if is_old_openai:

            async def logging_async_generator(async_gen: AsyncGenerator[_T]) -> AsyncGenerator[_T]:
                async for chunk in async_gen:
                    self._update_accumulator_streaming(accumulator, info_from_response, chunk)
                    yield chunk
                agen_final_processing_and_logging(accumulator, info_from_response)

            return logging_async_generator(response)
        else:
            from parea.types import OpenAIAsyncStreamWrapper

            return OpenAIAsyncStreamWrapper(
                response,
                accumulator,
                info_from_response,
                OpenAIWrapper._update_accumulator_streaming,
                agen_final_processing_and_logging,
            )

    @staticmethod
    def update_trace_data_from_stream_response(trace_id, model, accumulator, kwargs, first_token_timestamp):
        output = OpenAIWrapper._get_output(accumulator, model)
        trace_data.get()[trace_id].output = output
        output_tokens = _num_tokens_from_string(output, model)
        input_tokens = _calculate_input_tokens(
            kwargs.get("messages", []),
            kwargs.get("functions", []) or [d["function"] for d in kwargs.get("tools", [])],
            kwargs.get("function_call", "auto") or kwargs.get("tool_choice", "auto"),
            model,
        )
        trace_data.get()[trace_id].input_tokens = input_tokens
        trace_data.get()[trace_id].output_tokens = output_tokens
        trace_data.get()[trace_id].total_tokens = input_tokens + output_tokens
        trace_data.get()[trace_id].cost = _compute_cost(input_tokens, output_tokens, model)
        trace_data.get()[trace_id].time_to_first_token = (first_token_timestamp - datetime.fromisoformat(trace_data.get()[trace_id].start_timestamp)).total_seconds()

    @staticmethod
    def _kwargs_to_llm_configuration(kwargs):
        return _kwargs_to_llm_configuration(kwargs)

    @staticmethod
    def _get_output(result: Any, model: Optional[str] = None) -> str:
        PARSED_CHAT_COMPLETION_AVAILABLE = False
        try:
            from openai.types.chat import ParsedChatCompletionMessage

            PARSED_CHAT_COMPLETION_AVAILABLE = True
        except ImportError:
            ParsedChatCompletionMessage = None

        if isinstance(result, dict):
            result = convert_to_openai_object(
                {
                    "choices": [
                        {
                            "index": 0,
                            "message": result,
                        }
                    ],
                    "model": model or "model",
                }
            )
        response_message = result.choices[0].message
        if PARSED_CHAT_COMPLETION_AVAILABLE and isinstance(response_message, ParsedChatCompletionMessage):
            completion = json_dumps(response_message.parsed) if response_message.parsed else ""
        elif not response_message.get("content", None) if is_old_openai else not response_message.content:
            completion = OpenAIWrapper._format_function_call(response_message)
        else:
            completion = response_message.content
        return completion

    @staticmethod
    def _format_function_call(response_message) -> str:
        return _format_function_call(response_message)

    def convert_kwargs_to_cache_request(self, _args: Sequence[Any], kwargs: Dict[str, Any]) -> CacheRequest:
        return CacheRequest(
            configuration=self._kwargs_to_llm_configuration(kwargs),
        )

    @staticmethod
    def _convert_cache_to_response(_args: Sequence[Any], kwargs: Dict[str, Any], cache_response: TraceLog) -> OpenAIObject:
        content = cache_response.output
        message = {"role": "assistant", "content": None}
        try:
            function_or_tool_call = json.loads(content)
            if isinstance(function_or_tool_call, list) and len(function_or_tool_call) == 1:
                if "name" in function_or_tool_call[0] and "arguments" in function_or_tool_call[0]:
                    message["function_call"] = function_or_tool_call[0]
            elif isinstance(function_or_tool_call, list) and len(function_or_tool_call) > 1:
                tool_calls = []
                for idx, tool_call in enumerate(function_or_tool_call):
                    if "name" in tool_call and "arguments" in tool_call:
                        tool_calls.append({"id": idx, "function": tool_call, "type": "function"})
                message["tool_calls"] = tool_calls
            else:
                message["content"] = content
        except json.JSONDecodeError:
            message["content"] = content

        message_field = "delta" if kwargs.get("stream", False) else "message"

        return convert_to_openai_object(
            {
                "object": "chat.completion",
                "model": cache_response.configuration.model,
                "choices": [
                    {
                        "index": 0,
                        message_field: message,
                    }
                ],
                "usage": {
                    "prompt_tokens": cache_response.input_tokens,
                    "completion_tokens": cache_response.output_tokens,
                    "total_tokens": cache_response.total_tokens,
                },
            }
        )

    def convert_cache_to_response(self, _args: Sequence[Any], kwargs: Dict[str, Any], cache_response: TraceLog) -> Union[OpenAIObject, Iterator[OpenAIObject]]:
        response = self._convert_cache_to_response(_args, kwargs, cache_response)
        if kwargs.get("stream", False):
            return iter([response])
        else:
            return response

    def aconvert_cache_to_response(self, _args: Sequence[Any], kwargs: Dict[str, Any], cache_response: TraceLog) -> Union[OpenAIObject, AsyncIterator[OpenAIObject]]:
        response = self._convert_cache_to_response(_args, kwargs, cache_response)
        if kwargs.get("stream", False):

            def aiterator(iterable):
                async def gen():
                    for item in iterable:
                        yield item

                return gen()

            return aiterator([response])
        else:
            return response

    @staticmethod
    def should_use_gen_resolver(response: Any) -> bool:
        if is_old_openai:
            return isinstance(response, (Iterator, AsyncIterator))
        else:
            from openai import AsyncStream, Stream

            return isinstance(response, (Stream, AsyncStream))



from typing import Callable, Optional

import ast
import re

from parea.evals.utils import call_openai
from parea.schemas.log import Log

one_score_pattern = re.compile(r"\[\[(\d+\.?\d*)\]\]")
one_score_pattern_backup = re.compile(r"\[(\d+\.?\d*)\]")


def llm_grader_factory(model: str = "gpt-4", question_field: str = "question", is_azure: Optional[bool] = False) -> Callable[[Log], float]:
    """
    This factory creates an evaluation function that uses an LLM to grade the response of an LLM to a given question.
    It is based on the paper [Judging LLM-as-a-Judge with MT-Bench and Chatbot Arena](https://arxiv.org/abs/2306.05685)
    which introduces general-purpose zero-shot prompt to rate responses from an LLM to a given question on a scale from 1-10.
    They find that GPT-4's ratings agree as much with a human rater as a human annotator agrees with another one (>80%).
    Further, they observe that the agreement with a human annotator increases as the response rating gets clearer.
    Additionally, they investigated how much the evaluating LLM overestimated its responses and found that GPT-4 and
    Claude-1 were the only models that didn't overestimate themselves.

    Args:
        is_azure: Whether to use the Azure API. Defaults to False.
        model: The model which should be used for grading. Currently, only supports OpenAI chat models.
        question_field: The key name/field used for the question/query of the user. Defaults to "question".

    Returns:
        Callable[[Log], float]: A function that takes a log as input and returns a score between 0 and 1 which is the
        rating of the response on a scale from 1-10 divided by 10.
    """

    def llm_grader(log: Log) -> float:
        question = log.inputs[question_field]
        output = log.output
        rating_response = call_openai(
            model=model,
            messages=[
                {"role": "system", "content": "You are a helpful assistant."},
                {
                    "role": "user",
                    "content": f"[Instruction]\nPlease act as an impartial judge and evaluate the quality of the response "
                    f"provided by an AI assistant to the user question displayed below. Your evaluation should "
                    f"consider factors such as the helpfulness, relevance, accuracy, depth, creativity, and "
                    f"level of detail of the response. Begin your evaluation by providing a short explanation. "
                    f"Be as objective as possible. After providing your explanation, you must rate the response "
                    f'on a scale of 1 to 10 by strictly following this format: "[[rating]]", for example: '
                    f'"Rating: [[5]]".\n\n[Question]\n{question}\n\n[The Start of Assistant\'s Answer]'
                    f"\n{output}\n[The End of Assistant's Answer]",
                },
            ],
            temperature=0.0,
            is_azure=is_azure,
        )
        match = re.search(one_score_pattern, rating_response)
        if not match:
            match = re.search(one_score_pattern_backup, rating_response)

        if match:
            rating = ast.literal_eval(match.groups()[0])
        else:
            rating = 0

        return rating / 10.0

    return llm_grader


llm_grader_gpt4 = llm_grader_factory("gpt-4")
llm_grader_gpt3t = llm_grader_factory("gpt-3.5-turbo-16k")



from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple

import json
from functools import wraps

from dspy import Example, Module, Prediction
from wrapt import wrap_object

from parea import trace
from parea.schemas import EvaluationResult, Log
from parea.utils.trace_integrations.wrapt_utils import CopyableFunctionWrapper
from parea.utils.universal_encoder import json_dumps

_DSPY_MODULE_NAME = "dspy"
_DSP_MODULE_NAME = "dsp"


class DSPyInstrumentor:

    def instrument(self) -> None:
        # Instrument LM (language model) calls
        from dsp.modules.lm import LM
        from dspy import Predict

        language_model_classes = LM.__subclasses__()
        for lm in language_model_classes:
            wrap_object(
                module=_DSP_MODULE_NAME,
                name=lm.__name__ + ".basic_request",
                factory=CopyableFunctionWrapper,
                args=(_GeneralDSPyWrapper("request"),),
            )

        # Predict is a concrete (non-abstract) class that may be invoked
        # directly, but DSPy also has subclasses of Predict that override the
        # forward method. We instrument both the forward methods of the base
        # class and all subclasses.
        wrap_object(
            module=_DSPY_MODULE_NAME,
            name="Predict.forward",
            factory=CopyableFunctionWrapper,
            args=(_PredictForwardWrapper(),),
        )

        predict_subclasses = Predict.__subclasses__()
        for predict_subclass in predict_subclasses:
            wrap_object(
                module=_DSPY_MODULE_NAME,
                name=predict_subclass.__name__ + ".forward",
                factory=CopyableFunctionWrapper,
                args=(_PredictForwardWrapper(),),
            )

        wrap_object(
            module=_DSPY_MODULE_NAME,
            name="Retrieve.__call__",
            factory=CopyableFunctionWrapper,
            args=(_GeneralDSPyWrapper("forward"),),
        )

        wrap_object(
            module=_DSPY_MODULE_NAME,
            # At this time, dspy.Module does not have an abstract forward
            # method, but assumes that user-defined subclasses implement the
            # forward method and invokes that method using __call__.
            name="Module.__call__",
            factory=CopyableFunctionWrapper,
            args=(_GeneralDSPyWrapper("forward"),),
        )

        # At this time, there is no common parent class for retriever models as
        # there is for language models. We instrument the retriever models on a
        # case-by-case basis.
        wrap_object(
            module=_DSP_MODULE_NAME,
            name="ColBERTv2.__call__",
            factory=CopyableFunctionWrapper,
            args=(_GeneralDSPyWrapper("__call__"),),
        )


class _GeneralDSPyWrapper:
    def __init__(self, method_name: str):
        self._method_name = method_name

    def __call__(
        self,
        wrapped: Callable[..., Any],
        instance: Any,
        args: Tuple[type, Any],
        kwargs: Mapping[str, Any],
    ) -> Any:
        span_name = instance.__class__.__name__ + "." + self._method_name
        return trace(name=span_name)(wrapped)(*args, **kwargs)


class _PredictForwardWrapper:
    """
    A wrapper for the Predict class to have a chain span for each prediction
    """

    def __call__(
        self,
        wrapped: Callable[..., Any],
        instance: Any,
        args: Tuple[type, Any],
        kwargs: Mapping[str, Any],
    ) -> Any:
        from dspy import Predict

        # At this time, subclasses of Predict override the base class' forward
        # method and invoke the parent class' forward method from within the
        # overridden method. The forward method for both Predict and its
        # subclasses have been instrumented. To avoid creating duplicate spans
        # for a single invocation, we don't create a span for the base class'
        # forward method if the instance belongs to a proper subclass of Predict
        # with an overridden forward method.
        is_instance_of_predict_subclass = isinstance(instance, Predict) and (cls := instance.__class__) is not Predict
        has_overridden_forward_method = getattr(cls, "forward", None) is not getattr(Predict, "forward", None)
        wrapped_method_is_base_class_forward_method = wrapped.__qualname__ == Predict.forward.__qualname__
        if is_instance_of_predict_subclass and has_overridden_forward_method and wrapped_method_is_base_class_forward_method:
            return wrapped(*args, **kwargs)
        else:
            return trace(name=_get_predict_span_name(instance))(wrapped)(*args, **kwargs)


def _get_predict_span_name(instance: Any) -> str:
    """
    Gets the name for the Predict span, which are the composition of a Predict
    class or subclass and a user-defined signature. An example name would be
    "Predict(UserDefinedSignature).forward".
    """
    class_name = str(instance.__class__.__name__)
    if (signature := getattr(instance, "signature", None)) and (signature_name := _get_signature_name(signature)):
        return f"{class_name}({signature_name}).forward"
    return f"{class_name}.forward"


def _get_signature_name(signature: Any) -> Optional[str]:
    """
    A best-effort attempt to get the name of a signature.
    """
    if (
        # At the time of this writing, the __name__ attribute on signatures does
        # not return the user-defined class name, but __qualname__ does.
        qual_name := getattr(signature, "__qualname__", None)
    ) is None:
        return None
    return str(qual_name.split(".")[-1])


def convert_dspy_eval_to_parea(dspy_eval):
    """Given a DSPy evaluation function, convert it to a Parea evaluation function."""

    @wraps(dspy_eval)
    def wrapper(log: Log):
        inputs_dict = log.inputs or {}
        target_dict = json.loads(log.target) if log.target else {}
        output_dict = json.loads(log.output) if log.output else {}
        ex = Example(**inputs_dict, **target_dict).with_inputs(*inputs_dict.keys())
        pred = Prediction(**output_dict)
        score = dspy_eval(ex, pred)
        return EvaluationResult(score=score, name=dspy_eval.__name__)

    return wrapper


def convert_dspy_examples_to_parea_dicts(examples: List[Example]) -> List[Dict]:
    """Convert a list of DSPy examples to a list of dictionaries suitable for Parea."""
    converted_examples = []
    for example in examples:
        _dict = {k: v for k, v in example.inputs().items()}
        _target = {k: v for k, v in example.labels().items()}
        if _target:
            _dict["target"] = json_dumps(_target)
        converted_examples.append(_dict)
    return converted_examples


def attach_evals_to_module(module: Module, eval_funcs: List) -> Callable:
    """Attach evaluation functions to a DSPy module to apply them async after execution of module."""

    @trace(name=module.__class__.__name__, eval_funcs=[convert_dspy_eval_to_parea(f) for f in eval_funcs])
    def inner(*args, **kwargs):
        return module(*args, **kwargs)

    return inner



from typing import Any, AsyncGenerator, AsyncIterator, Callable, Dict, Generator, Iterator, List, Optional, Tuple

import contextvars
import inspect
import json
import logging
import os
import threading
import traceback
from collections import ChainMap
from datetime import datetime
from functools import wraps
from random import random

from parea.constants import PAREA_OS_ENV_EXPERIMENT_UUID, TURN_OFF_PAREA_EVAL_LOGGING
from parea.helpers import gen_trace_id, is_logging_disabled, timezone_aware_now
from parea.parea_logger import parea_logger
from parea.schemas import EvaluationResult
from parea.schemas.models import TraceLog, UpdateLog, UpdateTraceScenario
from parea.utils.universal_encoder import json_dumps

logger = logging.getLogger()


# Context variable to maintain the current trace context stack
trace_context = contextvars.ContextVar("trace_context", default=[])

# A dictionary to hold trace data for each trace
trace_data = contextvars.ContextVar("trace_data", default={})

# Context variable to maintain running evals in thread
thread_ids_running_evals = contextvars.ContextVar("thread_ids_running_evals", default=[])

# Add a counter variable to maintain the execution order
execution_order_counters = contextvars.ContextVar("execution_order_counters", default={})


def clear_trace_context():
    global trace_context, trace_data, thread_ids_running_evals, execution_order_counters
    trace_context = contextvars.ContextVar("trace_context", default=[])
    trace_data = contextvars.ContextVar("trace_data", default={})
    thread_ids_running_evals = contextvars.ContextVar("thread_ids_running_evals", default=[])
    execution_order_counters = contextvars.ContextVar("execution_order_counters", default={})


def log_in_thread(target_func: Callable, data: Dict[str, Any]):
    context = contextvars.copy_context()
    logging_thread = threading.Thread(target=context.run, args=(target_func,), kwargs=data)
    logging_thread.start()


def merge(old, new, key=None):
    if key == "error" and old:
        return json_dumps([old, new])
    if isinstance(old, dict) and isinstance(new, dict):
        return dict(ChainMap(new, old))
    if isinstance(old, list) and isinstance(new, list):
        return old + new
    return new


def check_multiple_return_values(func) -> bool:
    specs = inspect.getfullargspec(func)
    try:
        r = specs.annotations.get("return")
        if r and r.__origin__ == tuple:
            return len(r.__args__) > 1
    except Exception:
        return False


def make_output(result, islist) -> Optional[str]:
    if result is None:
        return result
    try:
        if islist:
            json_list = [json_dumps(r) for r in result]
            return json_dumps(json_list)
        elif isinstance(result, str):
            return result
        else:
            return json_dumps(result)
    except Exception as e:
        logger.debug(f"Error occurred making output with result: {result}. Error: {e}", exc_info=e)
        return str(result)


def get_current_trace_id() -> str:
    stack = trace_context.get()
    if stack:
        return stack[-1]
    return ""


def get_root_trace_id() -> str:
    stack = trace_context.get()
    if stack:
        return stack[0]
    return ""


def trace_insert(data: Dict[str, Any], trace_id: Optional[str] = None):
    """
    Insert data into the trace log for the current or specified trace id. Data should be a dictionary with keys that correspond to the fields of the TraceLog model.
    If the field already has an existing value that is extensible (dict, set, list, etc.), the new value will be merged with the existing value.
    Args:
        data: Keys can be one of: trace_name, end_user_identifier, metadata, tags, deployment_id, images, session_id
        trace_id: The trace id to insert the data into. If not provided, the current trace id will be used.
    """
    try:
        current_trace_id = trace_id or get_current_trace_id()
        current_trace_data: TraceLog = trace_data.get()[current_trace_id]
        if not current_trace_data:
            return
        for key, new_value in data.items():
            existing_value = current_trace_data.__getattribute__(key)
            current_trace_data.__setattr__(key, merge(existing_value, new_value, key) if existing_value else new_value)
    except Exception as e:
        logger.debug(f"Error occurred inserting data into trace log, {e}", exc_info=e)


def fill_trace_data(trace_id: str, data: Dict[str, Any], scenario: UpdateTraceScenario):
    try:
        if scenario == UpdateTraceScenario.RESULT:
            if not isinstance(data["result"], (Generator, AsyncGenerator, AsyncIterator, Iterator)):
                trace_data.get()[trace_id].output = make_output(data["result"], data.get("output_as_list", False))
            trace_data.get()[trace_id].status = "success"
            trace_data.get()[trace_id].evaluation_metric_names = data.get("eval_funcs_names")
        elif scenario == UpdateTraceScenario.ERROR:
            trace_data.get()[trace_id].error = data["error"]
            trace_data.get()[trace_id].status = "error"
        elif scenario == UpdateTraceScenario.CHAIN:
            parent_trace_id = data["parent_trace_id"]
            trace_data.get()[trace_id].parent_trace_id = parent_trace_id
            trace_data.get()[parent_trace_id].children.append(trace_id)
            if trace_data.get()[parent_trace_id].log_sample_rate != trace_data.get()[trace_id].log_sample_rate:
                trace_data.get()[trace_id].log_sample_rate = trace_data.get()[parent_trace_id].log_sample_rate
        elif scenario == UpdateTraceScenario.LANGCHAIN_CHILD:
            trace_data.get()[data["parent_trace_id"]].children.append(trace_id)
        elif scenario == UpdateTraceScenario.OPENAICONFIG:
            trace_data.get()[trace_id].configuration = data["configuration"]
            trace_data.get()[trace_id].output = data["output"]
            trace_data.get()[trace_id].input_tokens = data["input_tokens"]
            trace_data.get()[trace_id].output_tokens = data["output_tokens"]
            trace_data.get()[trace_id].total_tokens = data["total_tokens"]
            trace_data.get()[trace_id].cost = data["cost"]
        else:
            logger.debug(f"Error occurred filling trace data. Scenario not valid: {scenario}")
    except Exception as e:
        logger.debug(f"Error occurred filling trace data for trace id {trace_id}, {e}", exc_info=e)


def trace(
    name: Optional[str] = None,
    tags: Optional[List[str]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    end_user_identifier: Optional[str] = None,
    session_id: Optional[str] = None,
    eval_funcs_names: Optional[List[str]] = None,
    eval_funcs: Optional[List[Callable]] = None,
    access_output_of_func: Optional[Callable] = None,
    apply_eval_frac: float = 1.0,
    deployment_id: Optional[str] = None,
    log_omit_inputs: Optional[bool] = False,
    log_omit_outputs: Optional[bool] = False,
    overwrite_trace_id: Optional[str] = None,
    overwrite_inputs: Optional[Dict[str, Any]] = None,
    log_sample_rate: Optional[float] = 1.0,
    fn_transform_generator_outputs: Callable[[List[Any]], str] = None,
):
    def init_trace(func_name, _parea_target_field, args, kwargs, func) -> Tuple[str, datetime, contextvars.Token]:
        start_time = timezone_aware_now()
        trace_id = overwrite_trace_id or gen_trace_id()

        new_trace_context = trace_context.get() + [trace_id]
        token = trace_context.set(new_trace_context)

        if is_logging_disabled():
            return trace_id, start_time, token

        try:
            if overwrite_inputs is not None:
                inputs = overwrite_inputs
            else:
                sig = inspect.signature(func)
                parameters = sig.parameters

                inputs = {k: v for k, v in zip(parameters.keys(), args)}
                inputs.update(kwargs)

            # filter out any values which aren't JSON serializable
            for k, v in inputs.items():
                try:
                    json.dumps(v)
                except TypeError:
                    try:
                        inputs[k] = json_dumps(v)
                    except (TypeError, AttributeError):
                        # if we can't serialize the value, just convert it to a string
                        inputs[k] = str(v)

            depth = len(new_trace_context) - 1
            root_trace_id = new_trace_context[0]

            # Get the execution order counter for the current root trace
            counters = execution_order_counters.get()
            if root_trace_id not in counters:
                counters[root_trace_id] = 0
            execution_order = counters[root_trace_id]
            counters[root_trace_id] += 1

            parent_trace_id = new_trace_context[-2] if len(new_trace_context) > 1 else None
            parent_target = trace_data.get()[parent_trace_id].target if parent_trace_id else None

            trace_data.get()[trace_id] = TraceLog(
                trace_id=trace_id,
                parent_trace_id=trace_id,
                root_trace_id=root_trace_id,
                start_timestamp=start_time.isoformat(),
                trace_name=name or func_name,
                end_user_identifier=end_user_identifier,
                session_id=session_id,
                metadata=metadata,
                target=_parea_target_field if _parea_target_field is not None else parent_target,
                tags=tags,
                inputs={} if log_omit_inputs else inputs,
                experiment_uuid=os.environ.get(PAREA_OS_ENV_EXPERIMENT_UUID, None),
                apply_eval_frac=apply_eval_frac,
                deployment_id=deployment_id,
                depth=depth,
                execution_order=execution_order,
                log_sample_rate=log_sample_rate,
            )

            if parent_trace_id:
                fill_trace_data(trace_id, {"parent_trace_id": parent_trace_id}, UpdateTraceScenario.CHAIN)
        except Exception as e:
            logger.debug(f"Error occurred initializing trace for function {func_name}, {e}")

        return trace_id, start_time, token

    def cleanup_trace(trace_id: str, start_time: datetime, context_token: contextvars.Token):
        end_time = timezone_aware_now()
        trace_data.get()[trace_id].end_timestamp = end_time.isoformat()
        trace_data.get()[trace_id].latency = (end_time - start_time).total_seconds()

        output = trace_data.get()[trace_id].output
        if trace_data.get()[trace_id].status == "success" and output:
            output_for_eval_metrics = output
            if access_output_of_func:
                try:
                    output = json.loads(output)
                    output = access_output_of_func(output)
                    output_for_eval_metrics = json_dumps(output)
                except Exception as e:
                    logger.error(f"Error accessing output of func with output: {output}. Error: {e}", exc_info=e)
            trace_data.get()[trace_id].output_for_eval_metrics = output_for_eval_metrics

        thread_eval_funcs_then_log(trace_id, eval_funcs)
        trace_context.reset(context_token)

    def _handle_iterator_cleanup(items, trace_id, start_time, context_token):
        if fn_transform_generator_outputs:
            result = fn_transform_generator_outputs(items)
        elif all(isinstance(item, str) for item in items):
            result = "".join(items)
        else:
            result = ""
        if not is_logging_disabled() and not log_omit_outputs:
            fill_trace_data(trace_id, {"result": result}, UpdateTraceScenario.RESULT)

        cleanup_trace(trace_id, start_time, context_token)

    async def _wrap_async_iterator(iterator, trace_id, start_time, context_token):
        items = []
        try:
            async for item in iterator:
                items.append(item)
                yield item
        finally:
            _handle_iterator_cleanup(items, trace_id, start_time, context_token)

    def _wrap_sync_iterator(iterator, trace_id, start_time, context_token):
        items = []
        try:
            for item in iterator:
                items.append(item)
                yield item
        finally:
            _handle_iterator_cleanup(items, trace_id, start_time, context_token)

    def decorator(func):
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            _parea_target_field = kwargs.pop("_parea_target_field", None)
            trace_id, start_time, context_token = init_trace(func.__name__, _parea_target_field, args, kwargs, func)
            output_as_list = check_multiple_return_values(func)
            result = None
            try:
                result = await func(*args, **kwargs)
                if not is_logging_disabled() and not log_omit_outputs:
                    fill_trace_data(trace_id, {"result": result, "output_as_list": output_as_list, "eval_funcs_names": eval_funcs_names}, UpdateTraceScenario.RESULT)
            except Exception as e:
                logger.error(f"Error occurred in function {func.__name__}, {e}")
                fill_trace_data(trace_id, {"error": traceback.format_exc()}, UpdateTraceScenario.ERROR)
                raise e
            finally:
                try:
                    if inspect.isasyncgen(result):
                        return _wrap_async_iterator(result, trace_id, start_time, context_token)
                    else:
                        cleanup_trace(trace_id, start_time, context_token)
                        # to not swallow any exceptions
                        if result is not None:
                            return result
                except Exception as e:
                    logger.debug(f"Error occurred cleaning up trace for function {func.__name__}, {e}", exc_info=e)

        @wraps(func)
        def wrapper(*args, **kwargs):
            _parea_target_field = kwargs.pop("_parea_target_field", None)
            trace_id, start_time, context_token = init_trace(func.__name__, _parea_target_field, args, kwargs, func)
            output_as_list = check_multiple_return_values(func)
            result = None
            try:
                result = func(*args, **kwargs)
                if not is_logging_disabled() and not log_omit_outputs:
                    fill_trace_data(trace_id, {"result": result, "output_as_list": output_as_list, "eval_funcs_names": eval_funcs_names}, UpdateTraceScenario.RESULT)
            except Exception as e:
                logger.error(f"Error occurred in function {func.__name__}, {e}")
                fill_trace_data(trace_id, {"error": traceback.format_exc()}, UpdateTraceScenario.ERROR)
                raise e
            finally:
                try:
                    if inspect.isgenerator(result):
                        return _wrap_sync_iterator(result, trace_id, start_time, context_token)
                    else:
                        cleanup_trace(trace_id, start_time, context_token)
                        # to not swallow any exceptions
                        if result is not None:
                            return result
                except Exception as e:
                    logger.debug(f"Error occurred cleaning up trace for function {func.__name__}, {e}", exc_info=e)

        if inspect.iscoroutinefunction(func):
            return async_wrapper
        else:
            return wrapper

    if callable(name):
        func = name
        name = None
        return decorator(func)

    return decorator


def call_eval_funcs_then_log(trace_id: str, eval_funcs: List[Callable] = None):
    data = trace_data.get()[trace_id]

    if eval_funcs and data.status == "success" and random() <= data.apply_eval_frac:
        thread_ids_running_evals.get().append(trace_id)
        if data.target is None and (root_target := trace_data.get()[data.root_trace_id].target) is not None:
            data.target = root_target
        scores = []
        for func in eval_funcs:
            try:
                if TURN_OFF_PAREA_EVAL_LOGGING:
                    score = func(data)
                else:
                    score = trace()(func)(data)
                if isinstance(score, EvaluationResult):
                    scores.append(score)
                elif isinstance(score, list):
                    scores.extend(score)
                elif score is not None:
                    scores.append(EvaluationResult(name=func.__name__, score=score))
            except Exception as e:
                logger.error(f"Error occurred calling evaluation function '{func.__name__}', {e}", exc_info=e)
        trace_data.get()[trace_id].scores = scores
        thread_ids_running_evals.get().remove(trace_id)

    data_with_scores = trace_data.get()[trace_id]
    parea_logger.default_log(data=data_with_scores)


def logger_record_log(trace_id: str):
    log_in_thread(parea_logger.record_log, {"data": trace_data.get()[trace_id]})


def logger_all_possible(trace_id: str):
    log_in_thread(parea_logger.default_log, {"data": trace_data.get()[trace_id]})


def thread_eval_funcs_then_log(trace_id: str, eval_funcs: List[Callable] = None):
    log_in_thread(call_eval_funcs_then_log, {"trace_id": trace_id, "eval_funcs": eval_funcs})


def logger_update_record(data: UpdateLog):
    log_in_thread(parea_logger.update_log, {"data": data})



from typing import Callable, Optional

import re

from parea.evals.utils import call_openai
from parea.schemas.log import Log


def factual_inconsistency_scale_factory(article_field: Optional[str] = "article", model: Optional[str] = "gpt-4", is_azure: Optional[bool] = False) -> Callable[[Log], float]:
    """
    This factory creates an evaluation function that grades the factual consistency of a summary with the article on a scale from 1 to 10.
    It is based on the paper [ChatGPT as a Factual Inconsistency Evaluator for Text Summarization](https://arxiv.org/abs/2303.15621)
    which finds that using `gpt-3.5-turbo-0301` leads to a higher correlation with human expert judgment when grading
    the factuality of summaries on a scale from 1 to 10 than baseline methods such as SummaC and QuestEval.

    Args:
        article_field: The key name/field used for the content which should be summarized. Defaults to "article".
        model: The model which should be used for grading. Currently, only supports OpenAI chat models. Defaults to "gpt-4".
        is_azure: Whether to use the Azure API. Defaults to False.

    Returns:
        Callable[[Log], float]: A function that takes a log as input and returns a score between 0 and 1 indicating
        if the generated summary is factually consistent with the original text.
    """

    def factual_inconsistency_scale(log: Log) -> float:
        article = log.inputs[article_field]
        output = log.output
        prompt = f"""Score the following summary given the corresponding article with respect to consistency from 1 to 10. Note that consistency measures how much information included in the summary is present in the source article. 10 points indicate the summary contains only statements that are entailed by the source document.
    Article: {article}
    Summary: {output}
    Marks: """
        response = call_openai(
            model=model,
            messages=[
                {"role": "user", "content": prompt},
            ],
            temperature=0.0,
            is_azure=is_azure,
        )

        pattern = re.compile(r"\d+")
        match = pattern.search(response)
        if match:
            score = match.group()
        else:
            score = 0

        return float(score) / 10.0

    return factual_inconsistency_scale



from typing import Any, Dict, List, Optional, Union

import json
import math
from enum import Enum

from attrs import define, field


class TraceIntegrations(str, Enum):
    LANGCHAIN = "langchain"


class Role(str, Enum):
    user = "user"
    assistant = "assistant"
    system = "system"
    example_user = "example_user"
    example_assistant = "example_assistant"
    tool = "tool"


@define
class Message:
    content: str
    role: Role = Role.user

    def to_dict(self) -> Dict[str, str]:
        return {
            "content": self.content,
            "role": str(self.role),
        }


@define
class ModelParams:
    temp: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    frequency_penalty: Optional[float] = 0.0
    presence_penalty: Optional[float] = 0.0
    max_length: Optional[int] = None
    response_format: Optional[dict] = None
    safe_prompt: Optional[bool] = None


@define
class LLMInputs:
    """model choice should match an enabled model on the Parea platform."""

    model: Optional[str] = None
    provider: Optional[str] = None
    model_params: Optional[ModelParams] = ModelParams()
    messages: Optional[List[Message]] = None
    history: Optional[List[Message]] = None
    functions: Optional[List[Any]] = None
    function_call: Optional[Union[str, Dict[str, str]]] = None


@define
class Log:
    configuration: LLMInputs = LLMInputs()
    inputs: Optional[Dict[str, str]] = None
    output: Optional[str] = None
    target: Optional[str] = None
    latency: Optional[float] = 0.0
    time_to_first_token: Optional[float] = None
    input_tokens: Optional[int] = 0
    output_tokens: Optional[int] = 0
    total_tokens: Optional[int] = 0
    cost: Optional[float] = 0.0

    def convert_to_jsonl_row_for_finetuning(self) -> dict:
        """Converts the trace log to a row in the finetuning jsonl format."""
        jsonl_row = {"messages": [m.to_dict() for m in self.configuration.messages]}
        output = self.output
        try:
            tool_calls = json.loads(output)
            tools = self.configuration.functions
            # if 'arguments' is in the output, it was actually a function call
            if "arguments" in tool_calls:
                function_call = tool_calls[0] if isinstance(tool_calls, List) else tool_calls
                function_call["arguments"] = json.dumps(function_call["arguments"])
                assistant_response = {
                    "role": "assistant",
                    "function_call": function_call,
                }
                jsonl_row["functions"] = tools
            else:
                tool_calls = tool_calls if isinstance(tool_calls, List) else [tool_calls]
                for tool_call in tool_calls:
                    tool_call["arguments"] = json.dumps(tool_call["arguments"])
                assistant_response = {
                    "role": "assistant",
                    "tool_calls": tool_calls,
                }
                jsonl_row["tools"] = [{"type": "function", "function": tool} for tool in tools]
        except json.JSONDecodeError:
            assistant_response = {"role": "assistant", "content": output}
        jsonl_row["messages"].append(assistant_response)
        return jsonl_row


@define
class EvaluationResult:
    name: str
    score: float = field()
    reason: Optional[str] = None

    @score.validator
    def _check_score_is_finite(self, attribute, value):
        if not math.isfinite(value):
            raise ValueError(f"Score must be finite, got {value}")


@define
class EvaluatedLog(Log):
    scores: Optional[List[EvaluationResult]] = field(factory=list)

    def get_score(self, name: str) -> Optional[EvaluationResult]:
        for score in self.scores:
            if score.name == name:
                return score
        return None



from typing import Callable, Optional

import re

from parea.evals.utils import call_openai
from parea.schemas.log import Log


def likert_scale_factory(article_field: Optional[str] = "article", model: Optional[str] = "gpt-4", is_azure: Optional[bool] = False) -> Callable[[Log], float]:
    """
    This factory creates an evaluation function that grades the quality of a summary on a Likert scale from 1-5 along
    the dimensions of relevance, consistency, fluency, and coherence. It is based on the paper
    [Human-like Summarization Evaluation with ChatGPT](https://arxiv.org/abs/2304.02554) which finds that using `gpt-3.5-0301`
    leads to a highest correlation with human expert judgment when grading summaries on a Likert scale from 1-5 than baseline
    methods. Noteworthy is that [BARTScore](https://arxiv.org/abs/2106.11520) was very competitive to `gpt-3.5-0301`.

    Args:
        is_azure: Whether to use the Azure API. Defaults to False.
        article_field: The key name/field used for the content which should be summarized. Defaults to "article".
        model: The model which should be used for grading. Currently, only supports OpenAI chat models. Defaults to "gpt-4".

    Returns:
        Callable[[Log], float]: A function that takes a log as input and returns a score between 0 and 1 indicating
        the quality of the summary on a Likert scale from 1-5 along the dimensions of relevance, consistency, fluency, and coherence.
    """

    def likert_scale(log: Log) -> float:
        article = log.inputs[article_field]
        output = log.output
        prompt = f"""Evaluate the quality of summaries written for a news article. Rate each summary on four dimensions: relevance, consistency, fluency, and coherence. You should rate on a scale from 1 (worst) to 5 (best).

Definitions are as follows:
Relevance: The rating measures how well the summary captures the key points of the article. Consider whether all and only the important aspects are contained in the summary.
Consistency: The rating measures whether the facts in the summary are consistent with the facts in the original article. Consider whether the summary does reproduce all facts accurately and does not make up untrue information.
Fluency: This rating measures the quality of individual sentences, whether they are well-written and grammatically correct. Consider the quality of individual sentences.
Coherence: The rating measures the quality of all sentences collectively, to fit together and sound natural. Consider the quality of the summary as a whole.

The article and the summary are given below:
Article: {article}
Summary: {output}"""
        response = call_openai(
            model=model,
            messages=[
                {"role": "user", "content": prompt},
            ],
            temperature=0.0,
            is_azure=is_azure,
        )

        # extract the scores
        pattern = re.compile(r"\d+")
        matches = pattern.findall(response)
        if matches:
            scores = matches
        else:
            scores = [0, 0, 0, 0]

        # normalize the scores
        scores = [float(score) / 5.0 for score in scores]

        # average the scores
        return sum(scores) / len(scores)

    return likert_scale



from typing import Any, Dict, List, Optional

from parea.helpers import gen_random_name
from parea.schemas.models import CreateTestCase, CreateTestCaseCollection
from parea.utils.universal_encoder import json_dumps


def create_test_collection(data: List[Dict[str, Any]], name: Optional[str] = None) -> CreateTestCaseCollection:
    """Create a test case collection from a dictionary of test cases.
    Args:
        data: list of key-value pairs where keys represent input names.
            Each item in the list represent a test case row.
            Target and Tags are reserved keys. There can only be one target and tags key per dict item.
            If target is present it will represent the target/expected response for the inputs.
            If tags are present they must be a list of json_serializable values.
        name: A unique name for the test collection. If not provided a random name will be generated.

    Returns: CreateTestCaseCollection
    """
    if not name:
        name = gen_random_name()

    column_names = list({k for row in data for k in row.keys() if k not in ["target", "tags"]})
    test_cases = create_test_cases(data)

    return CreateTestCaseCollection(name=name, column_names=column_names, test_cases=test_cases)


def create_test_cases(data: List[Dict[str, Any]]) -> List[CreateTestCase]:
    """Create a list of test cases from a dictionary.
    Args:
        data: list of key-value pairs where keys represent input names.
            Each item in the list represent a test case row.
            Target and Tags are reserved keys. There can only be one target and tags key per dict item.
            If target is present it will represent the target/expected response for the inputs.
            If tags are present they must be a list of json_serializable values.

    Returns: List[CreateTestCase]
    """
    test_cases: List[CreateTestCase] = []
    for row in data:
        inputs: Dict[str, str] = {}
        target: Optional[str] = None
        tags: list = []
        for k, v in row.items():
            if k == "target":
                if target is not None:
                    print("There can only be one target key per test case. Only the first target will be used.")
                target = json_dumps(v)
            elif k == "tags":
                if not isinstance(v, list):
                    raise ValueError("Tags must be a list of json serializable values.")
                if tags:
                    print("There can only be one tags key per test case. Only the first set of tags will be used.")
                tags = [tag if isinstance(tag, str) else json_dumps(tag) for tag in v]
            else:
                inputs[k] = v if isinstance(v, str) else json_dumps(v)
        test_cases.append(CreateTestCase(inputs=inputs, target=target, tags=tags))

    return test_cases



from .factual_inconsistency_binary import factual_inconsistency_binary_factory
from .factual_inconsistency_scale import factual_inconsistency_scale_factory
from .likert_scale import likert_scale_factory



from typing import Any, Dict

from copy import copy, deepcopy

from wrapt import BoundFunctionWrapper, FunctionWrapper


class CopyableBoundFunctionWrapper(BoundFunctionWrapper):  # type: ignore
    """
    A bound function wrapper that can be copied and deep-copied. When used to
    wrap a class method, this allows the entire class to be copied and
    deep-copied.

    For reference, see
    https://github.com/GrahamDumpleton/wrapt/issues/86#issuecomment-426161271
    and
    https://wrapt.readthedocs.io/en/master/wrappers.html#custom-function-wrappers
    """

    def __copy__(self) -> "CopyableBoundFunctionWrapper":
        return CopyableBoundFunctionWrapper(copy(self.__wrapped__), self._self_instance, self._self_wrapper)

    def __deepcopy__(self, memo: Dict[Any, Any]) -> "CopyableBoundFunctionWrapper":
        return CopyableBoundFunctionWrapper(deepcopy(self.__wrapped__, memo), self._self_instance, self._self_wrapper)


class CopyableFunctionWrapper(FunctionWrapper):  # type: ignore
    """
    A function wrapper that can be copied and deep-copied. When used to wrap a
    class method, this allows the entire class to be copied and deep-copied.

    For reference, see
    https://github.com/GrahamDumpleton/wrapt/issues/86#issuecomment-426161271
    and
    https://wrapt.readthedocs.io/en/master/wrappers.html#custom-function-wrappers
    """

    __bound_function_wrapper__ = CopyableBoundFunctionWrapper

    def __copy__(self) -> "CopyableFunctionWrapper":
        return CopyableFunctionWrapper(copy(self.__wrapped__), self._self_wrapper)

    def __deepcopy__(self, memo: Dict[Any, Any]) -> "CopyableFunctionWrapper":
        return CopyableFunctionWrapper(deepcopy(self.__wrapped__, memo), self._self_wrapper)



from typing import Callable, List, Optional, Union

import json
import warnings

import openai
import pysbd
import tiktoken
from attrs import define
from openai import __version__ as openai_version

from parea.parea_logger import parea_logger
from parea.schemas import EvaluationResult
from parea.schemas.log import Log
from parea.utils.trace_utils import thread_ids_running_evals, trace_data

seg = pysbd.Segmenter(language="en", clean=False)


@define
class EvalFuncTuple:
    name: str
    func: Callable


def sent_tokenize(text: str) -> List[str]:
    """Split into sentences"""
    sentences = seg.segment(text)
    assert isinstance(sentences, list)
    return sentences


def safe_json_loads(s) -> dict:
    try:
        return json.loads(s)
    except ValueError as e:
        warnings.warn(f"Invalid json: {e}")

    return {}


def call_openai(
    messages, model, temperature=1.0, max_tokens=None, top_p=1.0, frequency_penalty=0.0, presence_penalty=0.0, response_format=None, n=1, is_azure=False
) -> Union[str, List[str]]:
    openai.api_type = "openai"
    if is_azure:
        from openai.lib.azure import AzureOpenAI

        openai.api_type = "azure"

        completion = AzureOpenAI().chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            response_format=response_format,
            n=n,
        )
        if n == 1:
            return completion.choices[0].message.content
        else:
            return [c.message.content for c in completion.choices]
    if openai_version.startswith("0."):
        completion = openai.ChatCompletion.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            n=n,
        )
        if n == 1:
            return completion.choices[0].message["content"]
        else:
            return [c.message["content"] for c in completion.choices]
    else:
        completion = openai.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            response_format=response_format,
            n=n,
        )
        if n == 1:
            return completion.choices[0].message.content
        else:
            return [c.message.content for c in completion.choices]


def embed(model, input, is_azure=False) -> List[float]:
    openai.api_type = "openai"
    if is_azure:
        from openai.lib.azure import AzureOpenAI

        openai.api_type = "azure"
        return AzureOpenAI().embeddings.create(model=model, input=input, encoding_format="float").data[0].embedding
    if openai_version.startswith("0."):
        return openai.Embedding.create(model=model, input=input, encoding_format="float").data[0]["embedding"]
    else:
        return openai.embeddings.create(model=model, input=input, encoding_format="float").data[0].embedding


def dcg(y_true, ranking):
    """Discounted cumulative gain (DCG) at rank k."""
    import numpy as np

    y_true = np.asarray(y_true)
    ranking = np.asarray(ranking)
    rel = y_true[ranking]
    gains = 2**rel - 1
    discounts = np.log2(np.arange(len(ranking)) + 2)
    return np.sum(gains / discounts)


def ndcg(y_true, ranking):
    """Normalized discounted cumulative gain (NDCG) at rank k"""
    import numpy as np

    k = len(ranking)
    best_ranking = np.argsort(y_true)[::-1]
    best = dcg(y_true, best_ranking[:k])
    return dcg(y_true, ranking) / best


# note name is extra odd to make sure that skip_decorator_if_func_in_stack works in 99.9% of cases
def _make_evaluations(trace_id: str, log: Log, eval_funcs: List[EvalFuncTuple], verbose: bool = False, sync: bool = False):
    scores = []
    for eval in eval_funcs:
        try:
            result = eval.func(log)
        except Exception as e:
            print(f"Error occurred calling evaluation function '{eval.func.__name__}', {e}")
            continue
        if isinstance(result, EvaluationResult):
            scores.append(result)
        elif isinstance(result, list):
            scores.extend(result)
        elif result is not None:
            scores.append(EvaluationResult(name=eval.name, score=result))

    trace_data.get()[trace_id].scores = scores
    trace_data.get()[trace_id].target = log.target
    data_with_scores = trace_data.get()[trace_id]
    thread_ids_running_evals.get().remove(trace_id)
    parea_logger.default_log(data=data_with_scores)
    if verbose:
        print("###Eval Results###")
        for score in scores:
            print(score)
        print(f"View trace at: https://app.parea.ai/logs/detailed/{trace_id} \n")
    if sync:
        return scores


def run_evals_in_thread_and_log(trace_id: str, log: Log, eval_funcs: List[EvalFuncTuple], verbose: bool = False):
    import threading

    logging_thread = threading.Thread(
        target=_make_evaluations,
        kwargs={"trace_id": trace_id, "log": log, "eval_funcs": eval_funcs, "verbose": verbose},
    )
    logging_thread.start()


def run_evals_synchronous(trace_id: str, log: Log, eval_funcs: List[EvalFuncTuple], verbose: bool = False) -> List[EvaluationResult]:
    return _make_evaluations(trace_id, log, eval_funcs, verbose, True)


def get_tokens(model: str, text: str) -> List[int]:
    if not text:
        return []
    fallback_model = "cl100k_base"
    try:
        encoding = tiktoken.encoding_for_model(model or fallback_model)
    except KeyError:
        encoding = tiktoken.get_encoding(fallback_model)
    try:
        return encoding.encode(text)
    except Exception as e:
        print(f"Error encoding text: {e}")
        return []


def get_context(log: Log, context_fields: Optional[List[str]] = None, as_list: bool = False) -> str:
    if context_fields:
        context_list = [log.inputs[context_field] for context_field in context_fields]
        return context_list if as_list else "\n".join(context_list)
    else:
        context = log.output
        try:
            loaded_context = json.loads(log.output)
            if isinstance(loaded_context, list):
                return loaded_context if as_list else "\n".join(loaded_context)
        except json.JSONDecodeError:
            pass
        return [context] if as_list else context



from typing import Any, AsyncIterable, Callable, Dict, List, Optional

import asyncio
import json
import os
import time
from functools import wraps
from importlib import metadata as importlib_metadata

import httpx
from dotenv import load_dotenv

load_dotenv()

MAX_RETRIES = 8
BACKOFF_FACTOR = 0.5


def retry_on_502(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    A decorator to retry a function or coroutine on encountering a 502 error.
    Parameters:
        - func: The function or coroutine to be decorated.
    Returns:
        - A wrapper function that incorporates retry logic.
    """

    @wraps(func)
    async def async_wrapper(*args, **kwargs):
        for retry in range(MAX_RETRIES):
            try:
                return await func(*args, **kwargs)
            except httpx.HTTPError as e:
                if not _should_retry(e, retry):
                    raise
                await asyncio.sleep(BACKOFF_FACTOR * (2**retry))

    @wraps(func)
    def sync_wrapper(*args, **kwargs):
        for retry in range(MAX_RETRIES):
            try:
                return func(*args, **kwargs)
            except httpx.HTTPError as e:
                if not _should_retry(e, retry):
                    raise
                time.sleep(BACKOFF_FACTOR * (2**retry))

    def _should_retry(error, current_retry):
        """Determines if the function should retry on error."""
        is_502_error = isinstance(error, httpx.HTTPStatusError) and error.response.status_code == 502
        is_last_retry = current_retry == MAX_RETRIES - 1
        return not is_last_retry and (isinstance(error, (httpx.ConnectError, httpx.ReadError, httpx.RemoteProtocolError)) or is_502_error)

    if asyncio.iscoroutinefunction(func):
        return async_wrapper
    else:
        return sync_wrapper


class HTTPClient:
    _instance = None
    base_url = os.getenv("PAREA_BASE_URL", "https://parea-ai-backend-us-9ac16cdbc7a7b006.onporter.run/api/parea/v1")
    api_key = None
    integrations: List[str] = []

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.sync_client = httpx.Client(base_url=cls.base_url, timeout=60 * 3.0)
            cls._instance.async_client = httpx.AsyncClient(base_url=cls.base_url, timeout=60 * 3.0)
        return cls._instance

    def set_api_key(self, api_key: str):
        self.api_key = api_key

    def add_integration(self, integration: str):
        if integration not in self.integrations:
            self.integrations.append(integration)

    def _get_headers(self, api_key: Optional[str] = None) -> Dict[str, str]:
        headers = {
            "x-api-key": self.api_key or api_key,
            "x-sdk-version": get_version(),
            "x-sdk-language": "python",
        }
        if self.integrations:
            headers["x-sdk-integrations"] = ",".join(self.integrations)
        return headers

    @retry_on_502
    def request(
        self,
        method: str,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        api_key: Optional[str] = None,
    ) -> httpx.Response:
        """
        Makes an HTTP request to the specified endpoint.
        """
        headers = self._get_headers(api_key=api_key)
        try:
            response = self.sync_client.request(method, endpoint, json=data, headers=headers, params=params)
            response.raise_for_status()
            return response
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 422:
                # update the error message to include the validation errors
                e.args = (f"{e.args[0]}: {e.response.json()}",)
            raise

    @retry_on_502
    async def request_async(
        self,
        method: str,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        api_key: Optional[str] = None,
    ) -> httpx.Response:
        """
        Makes an asynchronous HTTP request to the specified endpoint.
        """
        headers = self._get_headers(api_key=api_key)
        try:
            response = await self.async_client.request(method, endpoint, json=data, headers=headers, params=params)
            response.raise_for_status()
            return response
        except httpx.HTTPStatusError as e:
            print(f"HTTP Error {e.response.status_code} for {e.request.url}: {e.response.text}")
            raise

    @retry_on_502
    def stream_request(
        self,
        method: str,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        api_key: Optional[str] = None,
        chunk_size: Optional[int] = None,
    ):
        """
        Makes a streaming HTTP request to the specified endpoint, yielding chunks of data.
        """
        headers = self._get_headers(api_key=api_key)
        try:
            with self.sync_client.stream(method, endpoint, json=data, headers=headers, params=params, timeout=None) as response:
                response.raise_for_status()
                for chunk in response.iter_bytes(chunk_size):
                    yield parse_event_data(chunk)
        except httpx.HTTPStatusError as e:
            print(f"HTTP Error {e.response.status_code} for {e.request.url}: {e.response.text}")
            raise

    @retry_on_502
    async def stream_request_async(
        self,
        method: str,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        api_key: Optional[str] = None,
        chunk_size: Optional[int] = None,
    ) -> AsyncIterable[bytes]:
        """
        Makes an asynchronous streaming HTTP request to the specified endpoint, yielding chunks of data.
        """
        headers = self._get_headers(api_key=api_key)
        try:
            async with self.async_client.stream(method, endpoint, json=data, headers=headers, params=params, timeout=None) as response:
                response.raise_for_status()
                async for chunk in response.aiter_bytes(chunk_size):
                    yield parse_event_data(chunk)
        except httpx.HTTPStatusError as e:
            print(f"HTTP Error {e.response.status_code} for {e.request.url}: {e.response.text}")
            raise

    def close(self):
        """
        Closes the synchronous HTTP client.
        """
        self.sync_client.close()

    async def close_async(self):
        """
        Closes the asynchronous HTTP client.
        """
        await self.async_client.aclose()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close_async()


def parse_event_data(byte_data):
    decoded_data = byte_data.decode("utf-8")

    try:
        json_str = [line for line in decoded_data.split("\r\n") if line.startswith("data:")][0]
        json_str = json_str.replace("data: ", "")
        if json_str.startswith("ID_START"):
            return ""

        data_dict = json.loads(json_str)
        return data_dict["chunk"]
    except Exception as e:
        print(f"Error parsing event data: {e}")
        return None


def get_version() -> str:
    try:
        return importlib_metadata.version("parea-ai")
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"



from typing import Callable, Optional

from parea.evals.utils import call_openai
from parea.schemas.log import Log


def factual_inconsistency_binary_factory(article_field: Optional[str] = "article", model: Optional[str] = "gpt-4", is_azure: Optional[bool] = False) -> Callable[[Log], float]:
    """
    This factory creates an evaluation function that classifies if a summary is factually inconsistent with the original text.
    It is based on the paper [ChatGPT as a Factual Inconsistency Evaluator for Text Summarization](https://arxiv.org/abs/2303.15621)
    which suggests using an LLM to assess the factuality of a summary by measuring how consistent the summary is with
    the original text, posed as a binary classification. They find that `gpt-3.5-turbo-0301` outperforms
    baseline methods such as SummaC and QuestEval when identifying factually inconsistent summaries.

    Args:
        article_field: The key name/field used for the content which should be summarized. Defaults to "article".
        model: The model which should be used for grading. Currently, only supports OpenAI chat models. Defaults to "gpt-4".
        is_azure: Whether to use the Azure API. Defaults to False.

    Returns:
        Callable[[Log], float]: A function that takes a log as input and returns a score between 0 and 1 indicating
        if the generated summary is factually consistent with the original text.
    """

    def factual_inconsistency_binary(log: Log) -> float:
        article = log.inputs[article_field]
        output = log.output
        prompt = f"""Decide if the following summary is consistent with the corresponding article. Note that consistency means all information in the summary is supported by the article.
    Article: {article}
    Summary: {output}
    Explain your reasoning step by step then answer (yes or no) the question:"""
        response = call_openai(
            model=model,
            messages=[
                {"role": "user", "content": prompt},
            ],
            temperature=0.0,
            is_azure=is_azure,
        )
        return float("yes" in response.lower())

    return factual_inconsistency_binary



from typing import List

import argparse
import csv
import os
import sys
import traceback
from importlib import util

from .experiment import _experiments


def load_from_path(module_path):
    # Ensure the directory of user-provided script is in the system path
    dir_name = os.path.dirname(module_path)
    if dir_name not in sys.path:
        sys.path.insert(0, dir_name)

    module_name = os.path.basename(module_path)
    spec = util.spec_from_file_location(module_name, module_path)
    module = util.module_from_spec(spec)
    spec.loader.exec_module(module)

    if spec.name not in sys.modules:
        sys.modules[spec.name] = module


def read_input_file(file_path) -> List[dict]:
    with open(file_path) as file:
        reader = csv.DictReader(file)
        inputs = list(reader)
    return inputs


def experiment(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("file", help="Path to the experiment", type=str)
    parser.add_argument("--run_name", help="Name of the experiment run", type=str, default=None)

    parsed_args = parser.parse_args(args)

    try:
        load_from_path(parsed_args.file)
    except Exception as e:
        print(f"Error loading function: {e}\n", file=sys.stderr)
        traceback.print_exc()
        sys.exit(1)

    for _experiment in _experiments:
        _experiment.run(parsed_args.run_name)



from typing import Any, AsyncGenerator, Generator

import json
from collections import defaultdict

from parea.constants import CHUNK_DONE_SENTINEL
from parea.utils.trace_utils import get_current_trace_id
from parea.utils.universal_encoder import json_dumps
from parea.wrapper.utils import convert_openai_raw_stream_to_log


def process_stream_and_yield(response, data: dict) -> Generator:
    trace_id = get_current_trace_id()
    accumulated_content = []
    accumulated_tools = defaultdict(lambda: {"function": {"arguments": [], "name": ""}})

    for chunk in response.iter_lines():
        format_and_accumulate_streaming_chunk(trace_id, accumulated_content, accumulated_tools, data, chunk)
        yield chunk


async def aprocess_stream_and_yield(response, data: dict) -> AsyncGenerator:
    trace_id = get_current_trace_id()
    accumulated_content = []
    accumulated_tools = defaultdict(lambda: {"function": {"arguments": [], "name": ""}})

    async for chunk in response.aiter_lines():
        format_and_accumulate_streaming_chunk(trace_id, accumulated_content, accumulated_tools, data, chunk)
        yield chunk


def format_and_accumulate_streaming_chunk(trace_id: str, accumulated_content: list, accumulated_tools: dict, data: dict, chunk: Any) -> None:
    from openai.types.chat import ChatCompletionChunk

    try:
        chunk = chunk.decode("utf-8")
    except AttributeError:
        pass
    if chunk == CHUNK_DONE_SENTINEL:
        # when done send accumulated content to be logged in background thread
        convert_openai_raw_stream_to_log(accumulated_content, accumulated_tools, data, trace_id)
    else:
        chunk_data = raw_chunk_to_chat_completion_chunk(chunk)
        if isinstance(chunk_data, ChatCompletionChunk):
            for choice in chunk_data.choices or []:
                delta = choice.delta

                if delta.content:
                    accumulated_content.append(delta.content)

                if delta.function_call:
                    accumulated_tools[0]["function"]["name"] = delta.function_call.name or accumulated_tools[0]["function"]["name"]
                    if delta.function_call.arguments:
                        accumulated_tools[0]["function"]["arguments"].append(delta.function_call.arguments)

                for tool_call in delta.tool_calls or []:
                    tool_id = tool_call.index
                    accumulated_tools[tool_id]["function"]["name"] = tool_call.function.name or accumulated_tools[tool_id]["function"]["name"]
                    if tool_call.function.arguments:
                        accumulated_tools[tool_id]["function"]["arguments"].append(tool_call.function.arguments)


def raw_chunk_to_chat_completion_chunk(chunk: str):
    from openai.types.chat import ChatCompletionChunk

    try:
        return ChatCompletionChunk(**json.loads(chunk[6:].strip()))
    except json.JSONDecodeError:
        return chunk


def get_formatted_openai_response(r):
    # helper function to format the response from OpenAI
    if r["choices"][0]["message"].get("content"):
        return r["choices"][0]["message"]["content"].strip()
    elif r["choices"][0]["message"].get("function_call"):
        function_call = r["choices"][0]["message"]["function_call"]
        formatted_function_call = {
            "name": function_call["name"],
            "arguments": json.loads(function_call["arguments"]),
        }
        return json_dumps(formatted_function_call, indent=4)
    elif r["choices"][0]["message"].get("tool_calls"):
        formatted_tool_calls = []
        tool_calls = r["choices"][0]["message"]["tool_calls"]
        for tool_call in tool_calls:
            formatted_tool_call = {
                "name": tool_call["function"]["name"],
                "arguments": json.loads(tool_call["function"]["arguments"]),
            }
            formatted_tool_calls.append(formatted_tool_call)
        return json_dumps(formatted_tool_calls, indent=4)
    return json_dumps(r, indent=4)



from typing import Callable, List, Optional

from parea.evals.utils import call_openai, get_context, ndcg
from parea.schemas.log import Log


def context_ranking_listwise_factory(
    question_field: str = "question",
    context_fields: Optional[List[str]] = None,
    ranking_measurement="ndcg",
    n_contexts_to_rank=10,
    model: Optional[str] = "gpt-3.5-turbo-16k",
    is_azure: Optional[bool] = False,
) -> Callable[[Log], float]:
    """
    This factory creates an evaluation function that measures how well the retrieved contexts are ranked by relevancy to the given query
    by listwise estimation of the relevancy of every context to the query. It is based on the paper
    [Zero-Shot Listwise Document Reranking with a Large Language Model](https://arxiv.org/abs/2305.02156) which suggests using an LLM
    to rerank a list of contexts and use that to evaluate how well the contexts are ranked by relevancy to the given query.
    The authors used a progressive listwise reordering if the retrieved contexts don't fit into the context window of the LLM.

    Args:
        is_azure: Whether to use the Azure API. Defaults to False.
        model: The model which should be used for grading. Defaults to "gpt-3.5-turbo-16k".
        question_field (str): The name of the field in the log that contains the question. Defaults to "question".
        context_fields: An optional list of key names/fields used for the retrieved contexts in the input to function. If empty list or None, it will use the output field of the log as context. Defaults to None.
        ranking_measurement (str): The measurement to use for ranking. Defaults to "ndcg".
        n_contexts_to_rank (int): The number of contexts to rank listwise. Defaults to 10.

    Returns:
        Callable[[Log], float]: A function that takes a log as input and returns a score between 0 and 1 indicating
        how well the retrieved context is ranked by their relevancy.

    Raises:
        ValueError: If n_contexts_to_rank is less than 1.
    """
    if n_contexts_to_rank < 1:
        raise ValueError("n_contexts_to_rank must be at least 1.")

    def listwise_reranking(query: str, contexts: List[str]) -> List[int]:
        """Uses a LLM to listwise rerank the contexts. Returns the indices of the contexts in the order of their
        relevance (most relevant to least relevant)."""
        if len(contexts) == 0 or len(contexts) == 1:
            return list(range(len(contexts)))

        prompt = ""
        for i in range(len(contexts)):
            prompt += f"Passage{i + 1} = {contexts[i]}\n"

        prompt += f"""Query = {query}
        Passages = [Passage1, ..., Passage{len(contexts)}]
        Sort the Passages by their relevance to the Query.
        Sorted Passages = ["""

        sorted_list = call_openai(
            messages=[
                {
                    "role": "user",
                    "content": prompt,
                }
            ],
            model=model,
            temperature=0.0,
            is_azure=is_azure,
        )

        s = sorted_list.strip("[] ").replace(" ", "")
        number_strings = s.split(",")
        return [int(num) for num in number_strings if num.isdigit()]

    def progressive_reranking(query: str, contexts: List[str]) -> List[int]:
        """Returns the indices of the contexts in the order of their relevance (most relevant to least relevant)."""
        if len(contexts) <= n_contexts_to_rank:
            return listwise_reranking(query, contexts)

        window_size = n_contexts_to_rank
        window_step = n_contexts_to_rank // 2
        offset = len(contexts) - window_size

        indices = list(range(len(contexts)))

        while offset > 0:
            window_contexts = contexts[offset : offset + window_size]
            window_indices = indices[offset : offset + window_size]
            reranked_indices = listwise_reranking(query, window_contexts)
            contexts[offset : offset + window_size] = [window_contexts[i] for i in reranked_indices]
            indices[offset : offset + window_size] = [window_indices[i] for i in reranked_indices]

            offset -= window_step

        window_contexts = contexts[:window_size]
        window_indices = indices[:window_size]
        reranked_indices = listwise_reranking(query, window_contexts)
        contexts[:window_size] = [window_contexts[i] for i in reranked_indices]
        indices[:window_size] = [window_indices[i] for i in reranked_indices]

        return indices

    def context_ranking(log: Log) -> float:
        """Quantifies if the retrieved context is ranked by their relevancy by re-ranking the contexts."""
        question = log.inputs[question_field]
        contexts = get_context(log, context_fields, True)

        reranked_indices = progressive_reranking(question, contexts)

        if ranking_measurement == "ndcg":
            return ndcg(reranked_indices, list(range(len(contexts))))
        else:
            raise NotImplementedError

    return context_ranking



from typing import Any, Callable, List, Tuple

import contextvars
import functools
import inspect
import logging
import os
from datetime import datetime
from uuid import uuid4

from parea.cache.cache import Cache
from parea.constants import PAREA_OS_ENV_EXPERIMENT_UUID
from parea.evals.utils import _make_evaluations
from parea.helpers import is_logging_disabled, timezone_aware_now
from parea.schemas.models import TraceLog, UpdateLog, UpdateTraceScenario
from parea.utils.trace_utils import call_eval_funcs_then_log, execution_order_counters, fill_trace_data, logger_update_record, trace_context, trace_data
from parea.wrapper.utils import safe_format_template_to_prompt, skip_decorator_if_func_in_stack_for_evals

logger = logging.getLogger()


class Wrapper:
    def __init__(
        self,
        module: Any,
        func_names: List[str],
        resolver: Callable,
        gen_resolver: Callable,
        agen_resolver: Callable,
        should_use_gen_resolver: Callable,
        cache: Cache,
        convert_kwargs_to_cache_request: Callable,
        convert_cache_to_response: Callable,
        aconvert_cache_to_response: Callable,
        log: Callable,
        name: str = "LLM",
    ) -> None:
        self.resolver = resolver
        self.gen_resolver = gen_resolver
        self.agen_resolver = agen_resolver
        self.should_use_gen_resolver = should_use_gen_resolver
        self.log = log
        self.wrap_functions(module, func_names)
        self.cache = cache
        self.convert_kwargs_to_cache_request = convert_kwargs_to_cache_request
        self.convert_cache_to_response = convert_cache_to_response
        self.aconvert_cache_to_response = aconvert_cache_to_response
        self.name = name

    def wrap_functions(self, module: Any, func_names: List[str]):
        for func_name in func_names:
            func_name_parts = func_name.split(".")
            original = functools.reduce(getattr, func_name_parts, module)
            setattr(module if len(func_name_parts) == 1 else functools.reduce(getattr, func_name_parts[:-1], module), func_name_parts[-1], self._wrapped_func(original))

    def _wrapped_func(self, original_func: Callable) -> Callable:
        unwrapped_func = original_func
        is_wrapped_attribute_name = "_is_already_wrapped_by_parea"
        while hasattr(unwrapped_func, "__wrapped__"):
            if getattr(unwrapped_func, is_wrapped_attribute_name, False):
                return original_func
            unwrapped_func = unwrapped_func.__wrapped__
        if getattr(unwrapped_func, is_wrapped_attribute_name, False):
            return original_func

        wrapped_func = self._get_decorator(unwrapped_func, original_func)
        setattr(wrapped_func, is_wrapped_attribute_name, True)

        return wrapped_func

    def _get_decorator(self, unwrapped_func: Callable, original_func: Callable):
        if inspect.iscoroutinefunction(unwrapped_func):
            return self.async_decorator(original_func)
        else:
            return self.sync_decorator(original_func)

    def _init_trace(self, kwargs) -> Tuple[str, datetime, contextvars.Token]:
        start_time = timezone_aware_now()
        trace_id = str(uuid4())

        new_trace_context = trace_context.get() + [trace_id]
        token = trace_context.set(new_trace_context)

        if is_logging_disabled():
            return trace_id, start_time, token

        if template_inputs := kwargs.pop("template_inputs", None):
            for m in kwargs.get("messages", []):
                if isinstance(m, dict) and "content" in m:
                    if isinstance(m["content"], str):
                        m["content"] = safe_format_template_to_prompt(m["content"], **template_inputs)
                    elif isinstance(m["content"], list):
                        for i, item in enumerate(m["content"]):
                            if isinstance(item, dict) and "text" in item:
                                m["content"][i]["text"] = safe_format_template_to_prompt(item["text"], **template_inputs)

        depth = len(new_trace_context) - 1
        root_trace_id = new_trace_context[0]

        # Get the execution order counter for the current root trace
        counters = execution_order_counters.get()
        if root_trace_id not in counters:
            counters[root_trace_id] = 0
        execution_order = counters[root_trace_id]
        counters[root_trace_id] += 1

        try:
            trace_data.get()[trace_id] = TraceLog(
                trace_id=trace_id,
                parent_trace_id=root_trace_id,
                root_trace_id=root_trace_id,
                start_timestamp=start_time.isoformat(),
                trace_name=self.name,
                end_user_identifier=None,
                session_id=None,
                metadata=kwargs.pop("metadata", None),
                target=None,
                tags=None,
                inputs=template_inputs,
                experiment_uuid=os.getenv(PAREA_OS_ENV_EXPERIMENT_UUID, None),
                depth=depth,
                execution_order=execution_order,
            )

            parent_trace_id = new_trace_context[-2] if len(new_trace_context) > 1 else None
            if parent_trace_id:
                fill_trace_data(trace_id, {"parent_trace_id": parent_trace_id}, UpdateTraceScenario.CHAIN)
        except Exception as e:
            logger.debug(f"Error occurred initializing openai trace, {e}")

        return trace_id, start_time, token

    @skip_decorator_if_func_in_stack_for_evals(call_eval_funcs_then_log, _make_evaluations)
    def async_decorator(self, orig_func: Callable) -> Callable:
        @functools.wraps(orig_func)
        async def wrapper(*args, **kwargs):
            trace_id, start_time, context_token = self._init_trace(kwargs)
            response = None
            exception = None
            error = None
            cache_hit = False
            cache_key = self.convert_kwargs_to_cache_request(args, kwargs)
            try:
                if self.cache:
                    cache_result = await self.cache.aget(cache_key)
                    if cache_result is not None:
                        response = self.aconvert_cache_to_response(args, kwargs, cache_result)
                        cache_hit = True
                if response is None:
                    response = await orig_func(*args, **kwargs)
            except Exception as e:
                exception = e
                error = str(e)
                if self.cache:
                    await self.cache.ainvalidate(cache_key)
            finally:
                if exception is not None:
                    self._acleanup_trace(trace_id, start_time, error, cache_hit, args, kwargs, response, context_token)
                    raise exception
                else:
                    return self._acleanup_trace(trace_id, start_time, error, cache_hit, args, kwargs, response, context_token)

        return wrapper

    @skip_decorator_if_func_in_stack_for_evals(call_eval_funcs_then_log, _make_evaluations)
    def sync_decorator(self, orig_func: Callable) -> Callable:
        @functools.wraps(orig_func)
        def wrapper(*args, **kwargs):
            trace_id, start_time, context_token = self._init_trace(kwargs)
            response = None
            error = None
            cache_hit = False
            cache_key = self.convert_kwargs_to_cache_request(args, kwargs)
            exception = None
            try:
                if self.cache:
                    cache_result = self.cache.get(cache_key)
                    if cache_result is not None:
                        response = self.convert_cache_to_response(args, kwargs, cache_result)
                        cache_hit = True
                if response is None:
                    response = orig_func(*args, **kwargs)
            except Exception as e:
                exception = e
                error = str(e)
                if self.cache:
                    self.cache.invalidate(cache_key)
            finally:
                if exception is not None:
                    self._cleanup_trace(trace_id, start_time, error, cache_hit, args, kwargs, response, context_token)
                    raise exception
                else:
                    return self._cleanup_trace(trace_id, start_time, error, cache_hit, args, kwargs, response, context_token)

        return wrapper

    def _cleanup_trace_core(self, trace_id: str, start_time: datetime, error: str, cache_hit, args, kwargs, context_token: contextvars.Token):
        trace_data.get()[trace_id].cache_hit = cache_hit

        if error:
            trace_data.get()[trace_id].error = error
            trace_data.get()[trace_id].status = "error"
        else:
            trace_data.get()[trace_id].status = "success"

        def final_log():
            end_time = timezone_aware_now()
            trace_data.get()[trace_id].end_timestamp = end_time.isoformat()
            trace_data.get()[trace_id].latency = (end_time - start_time).total_seconds()

            if not error and self.cache:
                self.cache.set(self.convert_kwargs_to_cache_request(args, kwargs), trace_data.get()[trace_id])

            self.log(trace_id)
            try:
                trace_context_before_reset = trace_context.get()
                trace_context.reset(context_token)
                trace_context_after_reset = trace_context.get()
                if len(trace_context_after_reset) > len(trace_context_before_reset):
                    # this can happen if this is a streaming call and the LLM client got modified (e.g. instructor)
                    # so we need to manually reset the trace context to the previous state
                    trace_context.set(trace_context_before_reset)
                    # if the parent trace didn't have any output, we can also update the output
                    if (parent_trace_id := trace_data.get()[trace_id].parent_trace_id) is not None and not trace_data.get()[parent_trace_id].output:
                        logger_update_record(
                            UpdateLog(
                                trace_id=parent_trace_id,
                                field_name_to_value_map={"output": trace_data.get()[trace_id].output},
                                root_trace_id=trace_data.get()[parent_trace_id].root_trace_id,
                            )
                        )
            except IndexError:
                pass

        return final_log

    def _cleanup_trace(self, trace_id: str, start_time: datetime, error: str, cache_hit, args, kwargs, response, context_token: contextvars.Token):
        try:
            final_log = self._cleanup_trace_core(trace_id, start_time, error, cache_hit, args, kwargs, context_token)

            if self.should_use_gen_resolver(response):
                return self.gen_resolver(trace_id, args, kwargs, response, final_log)
            else:
                self.resolver(trace_id, args, kwargs, response)
                final_log()
                return response
        except Exception as e:
            logger.debug(f"Error occurred cleaning up openai trace, {e}")
            return response

    def _acleanup_trace(self, trace_id: str, start_time: datetime, error: str, cache_hit, args, kwargs, response, context_token: contextvars.Token):
        try:
            final_log = self._cleanup_trace_core(trace_id, start_time, error, cache_hit, args, kwargs, context_token)

            if self.should_use_gen_resolver(response):
                return self.agen_resolver(trace_id, args, kwargs, response, final_log)
            else:
                self.resolver(trace_id, args, kwargs, response)
                final_log()
                return response
        except Exception as e:
            logger.debug(f"Error occurred cleaning up openai trace, {e}")
            return response



from .answer_matches_target_llm_grader import answer_matches_target_llm_grader_factory
from .answer_matches_target_recall import answer_matches_target_recall
from .answer_relevancy import answer_relevancy_factory
from .levenshtein import levenshtein
from .llm_grader import llm_grader_factory, llm_grader_gpt3t, llm_grader_gpt4
from .lm_vs_lm import lm_vs_lm_factuality_factory, lm_vs_lm_factuality_gpt3t, lm_vs_lm_factuality_gpt4
from .self_check import self_check
from .semantic_similarity import semantic_similarity_factory, semantic_similarity_oai_3_large, semantic_similarity_oai_3_small, semantic_similarity_oai_ada_002



from typing import Any

import dataclasses
import datetime
import json
import logging
from decimal import Decimal
from enum import Enum
from uuid import UUID

import attrs
import openai
from pydantic import BaseModel

logger = logging.getLogger()


def is_dataclass_instance(obj):
    return dataclasses.is_dataclass(obj) and not isinstance(obj, type)


def is_attrs_instance(obj):
    return attrs.has(obj)


def is_numpy_instance(obj):
    try:
        import numpy as np
    except ImportError:
        np = None

    return np and isinstance(obj, np.ndarray)


def is_pandas_instance(obj):
    try:
        import pandas as pd
    except ImportError:
        pd = None

    return pd and isinstance(obj, pd.DataFrame)


def is_openai_stream_wrapper(obj):
    if openai.__version__.startswith("0."):
        return False
    else:
        from parea.types import OpenAIAsyncStreamWrapper, OpenAIStreamWrapper

        return isinstance(obj, (OpenAIStreamWrapper, OpenAIAsyncStreamWrapper))


class UniversalEncoder(json.JSONEncoder):
    """
    A JSON encoder that can handle additional types such as dataclasses, attrs, and more.
    """

    def handle_dspy_response(self, obj) -> Any:
        try:
            import dspy
        except ImportError:
            return None

        from dsp.templates.template_v3 import Template
        from dspy.primitives.example import Example

        if hasattr(obj, "completions") and hasattr(obj.completions, "_completions"):
            # multiple completions
            return obj.completions._completions
        elif hasattr(obj, "_asdict"):
            # convert namedtuples to dictionaries
            return obj._asdict()
        elif isinstance(obj, Example):
            # handles Prediction objects and other sub-classes of Example
            return getattr(obj, "_store", {})
        elif isinstance(obj, Template):
            return {
                "fields": [self.default(field) for field in obj.fields],
                "instructions": obj.instructions,
            }
        else:
            return None

    def handle_openai_not_given(self, obj) -> Any:
        try:
            from openai import NotGiven
        except ImportError:
            return None

        if isinstance(obj, NotGiven):
            return {"not_given": None}
        return None

    def default(self, obj: Any):
        if isinstance(obj, str):
            return obj
        elif isinstance(obj, set):
            return list(obj)
        elif isinstance(obj, Enum):
            try:
                return str(obj.value)
            except Exception:
                return str(obj)
        elif is_openai_stream_wrapper(obj):
            return str(obj)
        elif is_dataclass_instance(obj):
            return dataclasses.asdict(obj)
        elif is_attrs_instance(obj):
            return attrs.asdict(obj)
        elif isinstance(obj, BaseModel):
            try:
                return obj.model_dump()
            except Exception:
                return obj.dict()
        elif isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return obj.isoformat()
        elif isinstance(obj, datetime.timedelta):
            return obj.total_seconds()
        elif isinstance(obj, UUID):
            return str(obj)
        elif isinstance(obj, Decimal):
            return float(obj)
        elif is_numpy_instance(obj):
            return obj.tolist()
        elif is_pandas_instance(obj):
            return obj.to_dict(orient="records")
        elif dspy_response := self.handle_dspy_response(obj):
            return dspy_response
        elif is_openai_not_given := self.handle_openai_not_given(obj):
            return is_openai_not_given["not_given"]
        elif callable(obj):
            try:
                return f"<callable {obj.__name__}>"
            except AttributeError:
                return str(obj)
        elif isinstance(obj, bytes):
            return obj.decode(errors="ignore")
        else:
            return super().default(obj)


def json_dumps(obj, **kwargs) -> str:
    try:
        return json.dumps(obj, cls=UniversalEncoder, **kwargs) if not isinstance(obj, str) else obj
    except TypeError as e:
        logger.debug(f"Error serializing object: {obj} with error: {e}")
        return str(obj)



from typing import Callable, List, Optional

from parea.evals.utils import call_openai
from parea.schemas.log import Log


def answer_context_faithfulness_statement_level_factory(
    question_field: str = "question", context_fields: List[str] = ["context"], model: Optional[str] = "gpt-3.5-turbo-16k", is_azure: Optional[bool] = False
) -> Callable[[Log], float]:
    """
    This factory creates an evaluation function that measures the faithfulness of the generated answer to the given context
    by measuring how many statements from the generated answer can be inferred from the given context. It is based on the paper
    [RAGAS: Automated Evaluation of Retrieval Augmented Generation](https://arxiv.org/abs/2309.15217) which suggests using an LLM
    to create a list of all statements in the generated answer and assessing whether the given context supports each statement.

    Args:
        model: The model which should be used for grading. Defaults to "gpt-3.5-turbo-16k".
        is_azure: Whether to use the Azure API. Defaults to False.
        question_field: The key name/field used for the question/query of the user. Defaults to "question".
        context_fields: A list of key names/fields used for the retrieved contexts. Defaults to ["context"].

    Returns:
        Callable[[Log], float]: A function that takes a log as input and returns a score between 0 and 1 indicating
        if the retrieved context is relevant to the query.
    """

    def answer_context_faithfulness_statement_level(log: Log) -> float:
        """Quantifies how much the generated answer can be inferred from the retrieved context."""
        question = log.inputs[question_field]
        context = "\n".join(log.inputs[context_field] for context_field in context_fields)
        output = log.output

        completion = call_openai(
            model=model,
            messages=[
                {
                    "role": "user",
                    "content": f"""\
Given a question and answer, create one or more statements from each sentence in the given answer.
question: Who was  Albert Einstein and what is he best known for?
answer: He was a German-born theortical physicist, widely acknowledged to be one of the greatest and most influential physicists of all time. He was best known for developing the theory of relativity, he also made important contributions to the development of the theory of quantum mechanics.
statements:\nAlbert Einstein was born in Germany.\nAlbert Einstein was best known for his theory of relativity.
question: Cadmium Chloride is slightly soluble in this chemical, it is also called what?
answer: alcohol
statements:\nCadmium Chloride is slightly soluble in alcohol.
question: Were Shahul and Jithin of thee same nationality?
answer: They were from different countries.
statements:\nShahul and Jithin were from different countries.
question:{question}
answer: {output}
statements:\n""",
                }
            ],
            temperature=0.0,
            is_azure=is_azure,
        )
        statements = completion.strip().split("\n")
        statements_formatted = [f"{i+1}. {s.strip()}" for i, s in enumerate(statements)]

        verdicts = (
            call_openai(
                model=model,
                messages=[
                    {
                        "role": "user",
                        "content": f"""\
Prompt: Natural language inference
Consider the given context and following statements, then determine whether they are supported by the information present in the context.Provide a brief explanation for each statement before arriving at the verdict (Yes/No). Provide a final verdict for each statement in order at the end in the given format. Do not deviate from the specified format.

Context:\nJohn is a student at XYZ University. He is pursuing a degree in Computer Science. He is enrolled in several courses this semester, including Data Structures, Algorithms, and Database Management. John is a diligent student and spends a significant amount of time studying and completing assignments. He often stays late in the library to work on his projects.
statements:\n1. John is majoring in Biology.\n2. John is taking a course on Artificial Intelligence.\n3. John is a dedicated student.\n4. John has a part-time job.\n5. John is interested in computer programming.\n
Answer:
1. John is majoring in Biology.
Explanation: John's major is explicitly mentioned as Computer Science. There is no information suggesting he is majoring in Biology.  Verdict: No.
2. John is taking a course on Artificial Intelligence.
Explanation: The context mentions the courses John is currently enrolled in, and Artificial Intelligence is not mentioned. Therefore, it cannot be deduced that John is taking a course on AI. Verdict: No.
3. John is a dedicated student.
Explanation: The prompt states that he spends a significant amount of time studying and completing assignments. Additionally, it mentions that he often stays late in the library to work on his projects, which implies dedication. Verdict: Yes.
4. John has a part-time job.
Explanation: There is no information given in the context about John having a part-time job. Therefore, it cannot be deduced that John has a part-time job.  Verdict: No.
5. John is interested in computer programming.
Explanation: The context states that John is pursuing a degree in Computer Science, which implies an interest in computer programming. Verdict: Yes.
Final verdict for each statement in order: No. No. Yes. No. Yes.
context:\n{context}
statements:\n{statements_formatted}
Answer:
""",
                    }
                ],
                temperature=0.0,
                is_azure=is_azure,
            )
            .lower()
            .strip()
        )
        final_answer = "Final verdict for each statement in order:".lower()
        if final_answer in verdicts:
            verdicts = verdicts[verdicts.find(final_answer) + len(final_answer) :]
            yes_count = sum(0 if "yes" in answer else 1 for answer in verdicts.strip().split(".") if answer != "")
            return yes_count / len(statements_formatted)
        else:
            return max(0, output.count("verdict: no")) / len(statements_formatted)

    return answer_context_faithfulness_statement_level



from typing import Callable, Optional

import json

from parea.evals.utils import call_openai
from parea.schemas.log import Log


def goal_success_ratio_factory(
    use_output: Optional[bool] = False, message_field: Optional[str] = None, model: Optional[str] = "gpt-4", is_azure: Optional[bool] = False
) -> Callable[[Log], float]:
    """
    This factory creates an evaluation function that measures the success ratio of a goal-oriented conversation.
    Typically, a user interacts with a chatbot or AI assistant to achieve specific goals.
    This motivates to measure the quality of a chatbot by counting how many messages a user has to send before they reach their goal.
    One can further break this down by successful and unsuccessful goals to analyze user & LLM behavior.

    Concretely:
    1. Delineate the conversation into segments by splitting them by the goals the user wants to achieve.
    2. Assess if every goal has been reached.
    3. Calculate the average number of messages sent per segment.

    Args:
        is_azure: Whether to use Azure as the model. Defaults to False.
        model: The model which should be used for grading.
        use_output (Optional[bool], optional): Whether to use the output of the log to access the messages. Defaults to False.
        message_field (Optional[str], optional): The name of the field in the log that contains the messages.
            Defaults to None. If None, the messages are taken from the configuration attribute.

    # noqa: DAR201
    # noqa: DAR401
    """
    if use_output and message_field:
        raise ValueError("Only one of use_output and message_field can be set.")

    def goal_success_ratio(log: Log) -> float:
        """Returns the average amount of turns the user had to converse with the AI to reach their goals."""
        if use_output:
            output_list_dicts = json.loads(log.output)
            messages = [m for m in output_list_dicts]
        elif message_field:
            messages = [m for m in log.inputs[message_field]]
        else:
            messages = [m.to_dict() for m in log.configuration.messages]
            if log.output:
                messages.append({"role": "assistant", "content": log.output})

        # need to determine where does a new goal start
        conversation_segments = []
        start_index = 0
        end_index = 3
        while end_index < len(messages):
            user_follows_same_goal = call_openai(
                [
                    {
                        "role": "system",
                        "content": "Look at the conversation and to determine if the user is still following the same goal "
                        "or if they are following a new goal. If they are following the same goal, respond "
                        "SAME_GOAL. Otherwise, respond NEW_GOAL. In any case do not answer the user request!",
                    }
                ]
                + messages[start_index:end_index],
                model=model,
                is_azure=is_azure,
            )

            if user_follows_same_goal == "SAME_GOAL":
                end_index += 2
            else:
                conversation_segments.append(messages[start_index : end_index - 1])
                start_index = end_index - 1
                end_index += 2

        if start_index < len(messages):
            conversation_segments.append(messages[start_index:])

        # for now assume that the user reached their goal in every segment
        # return the average amount of turns the user had to converse with the AI to reach their goals
        return sum([2 / len(segment) for segment in conversation_segments]) / len(conversation_segments)

    return goal_success_ratio



from typing import Callable, Optional

from collections import Counter

from parea.evals.utils import get_tokens
from parea.schemas.log import Log


def answer_context_faithfulness_precision_factory(context_field: Optional[str] = "context") -> Callable[[Log], float]:
    """
        This factory creates an evaluation function that calculates the how many tokens in the generated answer are also present in the retrieved context.
    It is based on the paper [Evaluating Correctness and Faithfulness of Instruction-Following Models for Question Answering](https://arxiv.org/abs/2307.16877)
    which finds that this method only slightly lags behind GPT-4 and outperforms GPT-3.5-turbo (see Table 4 from the above paper).

        Args:
            context_field: The key name/field used for the retrieved context. Defaults to "context".

        Returns:
            Callable[[Log], float]: A function that takes a log as input and returns a score between 0 and 1 indicating
            how many tokens in the generated answer are also present in the retrieved context.
    """

    def answer_context_faithfulness_precision(log: Log) -> float:
        """Prop. of tokens in model generation which are also present in the retrieved context."""
        context = log.inputs[context_field]
        model = log.configuration.model

        context_tokens = get_tokens(model, context)
        output_tokens = get_tokens(model, log.output)

        if len(context_tokens) == 0:
            return 1.0
        elif len(output_tokens) == 0:
            return 0.0

        common_tokens = Counter(context_tokens) & Counter(output_tokens)
        num_common = sum(common_tokens.values())
        return num_common / len(output_tokens)

    return answer_context_faithfulness_precision



from typing import Any, Dict, List, Optional, Union

import logging
from uuid import UUID

from langchain_core.tracers import BaseTracer
from langchain_core.tracers.schemas import ChainRun, LLMRun, Run, ToolRun

from parea.helpers import is_logging_disabled
from parea.parea_logger import parea_logger
from parea.schemas import UpdateTraceScenario
from parea.schemas.log import TraceIntegrations
from parea.utils.trace_utils import fill_trace_data, get_current_trace_id, get_root_trace_id

logger = logging.getLogger()


class PareaAILangchainTracer(BaseTracer):
    parent_trace_id: UUID
    _parea_root_trace_id: str = None
    _parea_parent_trace_id: str = None
    _session_id: Optional[str] = None
    _tags: List[str] = []
    _metadata: Dict[str, Any] = {}
    _end_user_identifier: Optional[str] = None
    _deployment_id: Optional[str] = None
    _log_sample_rate: Optional[float] = 1.0

    def __init__(
        self,
        session_id: Optional[str] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        end_user_identifier: Optional[str] = None,
        deployment_id: Optional[str] = None,
        log_sample_rate: Optional[float] = 1.0,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._session_id = session_id
        self._end_user_identifier = end_user_identifier
        self._deployment_id = deployment_id
        self._log_sample_rate = log_sample_rate
        if tags:
            self._tags = tags
        if metadata:
            self._metadata = metadata

    def _persist_run(self, run: Union[Run, LLMRun, ChainRun, ToolRun]) -> None:
        if is_logging_disabled():
            return
        try:
            self.parent_trace_id = run.id
            # using .dict() since langchain Run class currently set to Pydantic v1
            data = run.dict()
            data["_parea_root_trace_id"] = self._parea_root_trace_id or None
            data["_session_id"] = self._session_id
            data["_tags"] = self._tags
            data["_metadata"] = self._metadata
            data["_end_user_identifier"] = self._end_user_identifier
            data["_deployment_id"] = self._deployment_id
            data["_log_sample_rate"] = self._log_sample_rate
            # check if run has an attribute execution order
            if (hasattr(run, "execution_order") and run.execution_order == 1) or run.parent_run_id is None:
                data["_parea_parent_trace_id"] = self._parea_parent_trace_id or None
            parea_logger.record_vendor_log(data, TraceIntegrations.LANGCHAIN)
        except Exception as e:
            logger.exception(f"Error occurred while logging langchain run: {e}", stack_info=True)

    def get_parent_trace_id(self) -> UUID:
        return self.parent_trace_id

    def _on_run_create(self, run: Run) -> None:
        if (hasattr(run, "execution_order") and run.execution_order == 1) or run.parent_run_id is None:
            # need to check if any traces already exist
            self._parea_root_trace_id = get_root_trace_id()
            if parent_trace_id := get_current_trace_id():
                self._parea_parent_trace_id = parent_trace_id
                fill_trace_data(str(run.id), {"parent_trace_id": parent_trace_id}, UpdateTraceScenario.LANGCHAIN_CHILD)

    def _on_llm_end(self, run: Run) -> None:
        self._persist_run(run)

    def _on_chain_end(self, run: Run) -> None:
        self._persist_run(run)



from typing import Union

from collections import Counter

from parea.evals.utils import get_tokens
from parea.schemas.log import Log


def answer_matches_target_recall(log: Log) -> Union[float, None]:
    """Prop. of tokens in target/reference answer which are also in model generation."""
    if (target := log.target) is None:
        return None
    output = log.output
    model = log.configuration.model

    target_tokens = get_tokens(model, target)
    output_tokens = get_tokens(model, output)

    if len(target_tokens) == 0:
        return 1.0
    common_tokens = Counter(target_tokens) & Counter(output_tokens)
    num_common = sum(common_tokens.values())
    return num_common / len(target_tokens)



import os
import subprocess

from parea.constants import PAREA_DVC_DIR, PAREA_DVC_METRICS_FILE, PAREA_DVC_YAML_FILE
from parea.utils.universal_encoder import json_dumps


def is_git_repo():
    try:
        subprocess.check_output(["git", "branch"], stderr=subprocess.STDOUT)
        return True
    except:
        return False


def save_results_to_dvc_if_init(experiment_name: str, metrics: dict):
    if not parea_dvc_initialized(only_check=True):
        return
    write_metrics_to_dvc(metrics)
    try:
        subprocess.run(["dvc", "exp", "save", "-n", experiment_name], check=True)
    except subprocess.CalledProcessError as e:
        print(f"Failed to save results to DVC: {e}")


def write_metrics_to_dvc(metrics: dict):
    git_root = subprocess.check_output(["git", "rev-parse", "--show-toplevel"], text=True, stderr=subprocess.STDOUT).strip()
    with open(os.path.join(git_root, PAREA_DVC_METRICS_FILE), "w") as f:
        f.write(json_dumps(metrics, indent=2))


def _check_has_been_committed(git_root: str, file: str) -> bool:
    output = subprocess.check_output(["git", "log", "--", file], cwd=git_root, text=True, stderr=subprocess.STDOUT)
    return output and len(output) > 0


def parea_dvc_initialized(only_check: bool) -> bool:
    print_fn = print if not only_check else lambda *args, **kwargs: None

    if not is_git_repo():
        print_fn("Git repository is not found. Please run `git init` to initialize a git repository.")
        return False

    git_root = subprocess.check_output(["git", "rev-parse", "--show-toplevel"], text=True, stderr=subprocess.STDOUT).strip()

    # make sure DVC is initialized
    if not os.path.exists(os.path.join(git_root, ".dvc")):
        print_fn("DVC is not initialized. Please run `dvc init` to initialize DVC.")
        return False

    # make sure dvc.yaml and metrics.json exist in .parea directory
    if not os.path.exists(os.path.join(git_root, PAREA_DVC_YAML_FILE)):
        if only_check:
            return False
        else:
            print_fn(f"{PAREA_DVC_YAML_FILE} is not found. Creating the file.")
            if not os.path.exists(os.path.join(git_root, PAREA_DVC_DIR)):
                os.mkdir(os.path.join(git_root, PAREA_DVC_DIR))
            with open(os.path.join(git_root, PAREA_DVC_YAML_FILE), "w") as f:
                f.write("metrics:\n  - metrics.json\n")
            subprocess.run(["git", "add", PAREA_DVC_YAML_FILE], cwd=git_root, check=True)
    if not os.path.exists(os.path.join(git_root, PAREA_DVC_METRICS_FILE)):
        if only_check:
            return False
        else:
            print_fn(f"{PAREA_DVC_METRICS_FILE} is not found. Creating the file.")
            if not os.path.exists(os.path.join(git_root, PAREA_DVC_DIR)):
                os.mkdir(os.path.join(git_root, PAREA_DVC_DIR))
            write_metrics_to_dvc({})
            subprocess.run(["git", "add", PAREA_DVC_METRICS_FILE], cwd=git_root, check=True)

    # make sure dvc.yaml and metrics.json are committed
    dvc_yaml_file_missing = not _check_has_been_committed(git_root, PAREA_DVC_YAML_FILE)
    dvc_metrics_file_missing = not _check_has_been_committed(git_root, PAREA_DVC_METRICS_FILE)
    if dvc_metrics_file_missing:
        print_fn(f"{PAREA_DVC_METRICS_FILE} is not committed. Please to commit the file to your git history.")
    if dvc_yaml_file_missing:
        print_fn(f"{PAREA_DVC_YAML_FILE} is not committed. Please to commit the file to your git history.")
    if dvc_metrics_file_missing or dvc_yaml_file_missing:
        return False

    print_fn("Parea's DVC integration is initialized.")
    return True



from typing import Callable, Optional

from parea.evals.utils import call_openai, embed
from parea.schemas.log import Log


def answer_relevancy_factory(
    question_field: str = "question",
    n_generations: int = 3,
    model: Optional[str] = "gpt-3.5-turbo-16k",
    embedding_model: str = "text-embedding-ada-002",
    is_azure: Optional[bool] = False,
) -> Callable[[Log], float]:
    """
    This factory creates an evaluation function that measures how relevant the generated response is to the given question.
    It is based on the paper [RAGAS: Automated Evaluation of Retrieval Augmented Generation](https://arxiv.org/abs/2309.15217)
    which suggests using an LLM to generate multiple questions that fit the generated answer and measure the cosine
    similarity of the generated questions with the original one.

    Args:
        is_azure: Whether to use the Azure API. Defaults to False.
        embedding_model: The model which should be used for embedding the text.
        model: The model which should be used for grading. Defaults to "gpt-3.5-turbo-16k".
        question_field: The key name/field used for the question/query of the user. Defaults to "question".
        n_generations: The number of questions which should be generated. Defaults to 3.

    Returns:
        Callable[[Log], float]: A function that takes a log as input and returns a score between 0 and 1 indicating
        if the generated response is relevant to the query.

    Raises:
        ImportError: If numpy is not installed.
    """
    try:
        import numpy as np
    except ImportError:
        raise ImportError("Please install numpy to use this metric.")

    def answer_relevancy(log: Log) -> float:
        """Quantifies how much the generated answer relates to the query."""
        question = log.inputs[question_field]
        output = log.output

        generated_questions = call_openai(
            model=model,
            messages=[
                {
                    "role": "user",
                    "content": f"""\
Generate question for the given answer.
Answer:\nThe PSLV-C56 mission is scheduled to be launched on Sunday, 30 July 2023 at 06:30 IST / 01:00 UTC. It will be launched from the Satish Dhawan Space Centre, Sriharikota, Andhra Pradesh, India
Question: When is the scheduled launch date and time for the PSLV-C56 mission, and where will it be launched from?

Answer: {output}
Question:""",
                }
            ],
            temperature=0.0,
            n=n_generations,
            is_azure=is_azure,
        )
        embedded_generated_questions = [embed(model=embedding_model, input=q, is_azure=is_azure) for q in generated_questions]
        embedded_question = embed(model=embedding_model, input=question, is_azure=is_azure)

        question_vec = np.asarray(embedded_question).reshape(1, -1)
        gen_question_vec = np.asarray(embedded_generated_questions)
        norm = np.linalg.norm(gen_question_vec, axis=1) * np.linalg.norm(question_vec, axis=1)
        return (np.dot(gen_question_vec, question_vec.T).reshape(-1) / norm).mean()

    return answer_relevancy
